import math
from abc import ABC, abstractmethod
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional, Tuple, Type, Union

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn import Dropout, Linear, Module

from .cache import KeyValueCache
from .embeddings import QueryKeyRotaryEmbeddings

_TORCH_SDP: ContextVar[bool] = ContextVar("torch_sdp", default=False)


@contextmanager
def enable_torch_sdp(use_torch_sdp: bool = True):
    """
    Enables Torch scaled dot product attention.

    Torch provides an implementation of scaled dot product attention that
    has many optimizations. For instance, in some scenarios Flash Attention
    is applied (`Dao et al., 2022`_). We do not use the Torch implementation
    by default, because it is still in beta.

    This context manager enables use of the Torch implementation of scaled
    dot product attention.

    .. code-block:: python

        with enable_torch_sdp():
            Y = bert_encoder(X)

    .. _Dao et al., 2022:  https://arxiv.org/abs/2205.14135
    """
    token = _TORCH_SDP.set(use_torch_sdp)
    try:
        yield
    finally:
        _TORCH_SDP.reset(token)


@dataclass
class AttentionMask:
    """
    Mask for attention calculation. Sequence elements for which the
    corresponding mask element is set to ``False`` are ignored during
    attention calculation.

    :param bool_mask:
        The boolean mask.
    """

    bool_mask: Tensor

    def __init__(self, bool_mask: Tensor):
        if bool_mask.dtype != torch.bool:
            raise ValueError("Expected the attention mask to be of dtype 'torch.bool'")

        if bool_mask.dim() == 2:
            batch_len, key_len = bool_mask.shape
            self.bool_mask = bool_mask.view([batch_len, 1, 1, key_len])
        elif bool_mask.dim() == 4:
            self.bool_mask = bool_mask
        else:
            raise ValueError(
                "The attention mask must be a tensor of shape [batch, key_len] or "
                "[batch_len, heads, query_len, key_len]"
            )

    def apply_logit_mask(self, input: Tensor) -> Tensor:
        """
        Use the attention mask to mask attention logits.

        :param input:
            Attention logits to apply the mask to.

            *Shape:* ``(batch_size, heads, query_len, key_len)``
        :returns:
            Logits with the attention mask applied.

            *Shape:* ``(batch_size, heads, query_len, key_len)``
        """
        blocked_value = torch.finfo(input.dtype).min
        return torch.where(self.bool_mask, input, blocked_value)

    def filter_batch_items(self, mask: Tensor) -> "AttentionMask":
        """
        Filter batch sequences from the attention mask.

        Sequences for which the mask is ``True`` are retained.

        :param mask:
            Mask of batch items to retain.

            *Shape:* ``(batch_size,)``
        :returns:
            Filtered mask.
        """
        if mask.ndim != 1:
            raise ValueError(
                f"Attention mask filter must be a 1D tensor, has {mask.ndim} dimensions."
            )
        if mask.size(0) != self.bool_mask.size(0):
            raise ValueError(
                f"Attention mask filter size ({mask.size(0)}) must match attention mask batch size ({self.bool_mask.size(0)})."
            )
        if mask.dtype != torch.bool:
            raise ValueError(
                f"Attention mask filter dtype must be bool, was: {mask.dtype}."
            )

        return AttentionMask(self.bool_mask[mask])

    def dim(self) -> int:
        """
        Return the number of dimensions in the mask.
        """
        return self.bool_mask.dim()

    def merge_mask(self, other: "AttentionMask") -> "AttentionMask":
        """
        Merge this attention mask with another attention mask.

        :param other:
            Attention mask to merge.
        :returns:
            Merged mask.
        """
        return AttentionMask(self.bool_mask.logical_and(other.bool_mask))

    def logit_mask(self, dtype: torch.dtype) -> Tensor:
        """
        Generate the logit mask for the given ``dtype``.

        Elements of the mask that are ``False`` are set to the
        minimum value of the ``dtype`` and the rest to zero. During
        softmax calculation, adding this mask to the logits will
        result in (near-)zero probabilities for the elements that
        are ``False``.

        :param dtype:
            Data type of the logit mask.
        :returns:
            Logit mask.
        """
        return (1.0 - self.bool_mask.to(dtype)) * torch.finfo(dtype).min

    def extend_length(self, count: int, fill_value: bool) -> "AttentionMask":
        """
        Extend the attention mask in the sequence length dimension
        by the given value.

        :param count:
            Number of new elements to insert.
        :param fill_value:
            Value to store in the new elements.
        :returns:
            Extended mask.
        """
        if count <= 0:
            raise ValueError(
                f"Attention mask sequence length extension requires a non-zero, positive value for 'count'"
            )

        ext = torch.full(
            self.shape[:3] + (count,),
            fill_value,
            device=self.device,
        )
        return AttentionMask(torch.cat([self.bool_mask, ext], dim=-1))

    @property
    def shape(self) -> torch.Size:
        """
        Return the shape of the mask.
        """
        return self.bool_mask.shape

    @property
    def device(self) -> torch.device:
        """
        Return the device of the mask.
        """
        return self.bool_mask.device


def create_causal_mask(query: Tensor, key: Tensor) -> AttentionMask:
    """
    Create a causal mask. A causal mask ensures that tokens
    cannot attend to succeeding tokens.

    :param query:
        Query to compute the causal mask for.

        *Shape:* ``(batch_size, heads, query_len, head_width)``
    :param key:
        Key to compute the causal mask for.

        *Shape:* ``(batch_size, heads, key_len, head_width)``
    :returns:
        The causal mask.

        *Shape:* ``(batch_size, heads, query_len, key_len)``
    """
    query_len = query.size(2)
    key_len = key.size(2)

    causal_mask = torch.tril(
        torch.ones((key_len, key_len), device=query.device, dtype=torch.bool),
    ).view(1, 1, key_len, key_len)
    return AttentionMask(causal_mask[:, :, key_len - query_len : key_len, :key_len])


class QkvSplit(ABC):
    """
    Query, key, value splitting strategies.

    After the input projection of the attention layer, we have an array with
    shape ``(batch_size, seq_len, n_heads * head_width)`` where ``n_heads`` is
    the sum of the number of query, key, and value heads. We need to split up
    the array into separate arrays for query, key, and value heads.

    Subclasses of this class implement different splitting strategies.
    """

    @abstractmethod
    def split(
        self,
        *,
        projection: Tensor,
        head_width: int,
        n_query_heads: int,
        n_key_value_heads: int,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Split attention heads in the projection in query, key, and value
        heads.

        :param projection:
            The fused query, key, value projection.

            *Shape:* ``(batch_size, seq_len, (n_query_heads + 2 * n_key_value_heads) * head_width)``
        :param head_width:
            Head width.
        :param n_query_heads:
            Number of query heads.
        :param n_key_value_heads:
            Number of key/value heads.
        :returns:
            Query, key, value tensors.

            *Shapes:*

            - Query: ``(batch_size, n_query_heads, seq_len, head_width)``

            - Key: ``(batch_size, n_key_value_heads, seq_len, head_width)``

            - Value: ``(batch_size, n_key_value_heads, seq_len, head_width)``
        """
        ...


class QkvSplitGroupedByHead(QkvSplit):
    """
    Default splitting strategy.

    First view the array as ``(batch_size, seq_len, n_heads, head_width)``,
    where ``n_heads`` is the sum of the number of query, key, and value
    heads. Then split up the array along the ``n_heads`` dimension.
    """

    def split(
        self,
        *,
        projection: Tensor,
        head_width: int,
        n_query_heads: int,
        n_key_value_heads: int,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        query, key, value = projection.split(
            [
                n_query_heads * head_width,
                n_key_value_heads * head_width,
                n_key_value_heads * head_width,
            ],
            dim=-1,
        )
        query = split_heads(query, n_query_heads)
        key = split_heads(key, n_key_value_heads)
        value = split_heads(value, n_key_value_heads)

        return query, key, value


class QkvSplitGroupedByKVHeads(QkvSplit):
    """
    Split up the projection in key/value-sized chunks.

    First view the array as ``(batch_size, seq_len, n_key_value_heads,
    n_chunks, head_width)``. Then split up the array along the ``n_chunks``
    dimension.
    """

    def split(
        self,
        *,
        projection: Tensor,
        head_width: int,
        n_query_heads: int,
        n_key_value_heads: int,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        batch_size, seq_len, _ = projection.shape

        grouped_projection = projection.view(
            batch_size,
            seq_len,
            -1,
            n_query_heads // n_key_value_heads + 2,
            head_width,
        )
        query, key, value = grouped_projection.split(
            [n_query_heads // n_key_value_heads, 1, 1], dim=-2
        )

        def permute_merge_groups(x, n_heads):
            return x.permute(0, 2, 3, 1, 4).reshape(
                batch_size, n_heads, seq_len, head_width
            )

        query = permute_merge_groups(query, n_query_heads)
        key = permute_merge_groups(key, n_key_value_heads)
        value = permute_merge_groups(value, n_key_value_heads)

        return query, key, value


class AttentionHeads:
    def __init__(
        self,
        *,
        n_query_heads: int,
        n_key_value_heads: int,
        qkv_split: QkvSplit,
    ):
        """
        Construct an attention head configuration. This constructor must
        not be used directly, its signature may change even within a semver
        version. Use the factory methods instead.

        :param n_query_heads:
            Number of query heads.
        :param n_key_value_heads:
            Number of key/value heads.
        :param qkv_split:
            How query, key, and value should be split when using
            :py:class:`~curated_transformers.layers.QkvMode.MERGED_SPLIT_AFTER`.
            Not used for other query, key, value modes.

        :meta private:
        """
        self._n_query_heads = n_query_heads
        self._n_key_value_heads = n_key_value_heads
        self._qkv_split = qkv_split

    @classmethod
    def uniform(
        cls,
        n_attention_heads: int,
        qkv_split: QkvSplit,
    ) -> "AttentionHeads":
        """
        Construct a head configuration where query, key, and value have the
        same number of attention heads.

        :param n_attention_heads:
            Number of attention heads.
        :param qkv_split:
            How query, key, and value should be split when using
            :py:class:`~curated_transformers.layers.QkvMode.MERGED_SPLIT_AFTER`.
            Not used for other query, key, value modes.
        """
        return cls(
            n_query_heads=n_attention_heads,
            n_key_value_heads=n_attention_heads,
            qkv_split=qkv_split,
        )

    @classmethod
    def multi_query(
        cls,
        n_query_heads: int,
        qkv_split: QkvSplit,
    ) -> "AttentionHeads":
        """
        Construct a multi-query attention configuration: key has one head,
        value has one head, query has ``n_query_heads`` heads
        (`Shazeer et al., 2019`_). The key head and the value head are
        broadcast to the shape of the query.

        .. _Shazeer et al., 2019: https://arxiv.org/abs/1911.02150

        :param n_query_heads:
            Number of query heads.
        :param qkv_split:
            How query, key, and value should be split when using
            :py:class:`~curated_transformers.layers.QkvMode.MERGED_SPLIT_AFTER`.
            Not used for other query, key, value modes.
        """
        return cls(
            n_query_heads=n_query_heads, n_key_value_heads=1, qkv_split=qkv_split
        )

    @classmethod
    def key_value_broadcast(
        cls,
        *,
        n_query_heads: int,
        n_key_value_heads: int,
        qkv_split: QkvSplit,
    ) -> "AttentionHeads":
        """
        Construct a head configuration where query has a larger number
        of heads than key and value. Key/value heads are broadcast to
        correspond to the number of query heads.

        :param n_query_heads:
            Number of attention heads. Must be a multiple of
            ``n_key_value_heads``.
        :param n_key_value_heads:
            Number of key and value heads.
        :param qkv_split:
            How query, key, and value should be split when using
            :py:class:`~curated_transformers.layers.QkvMode.MERGED_SPLIT_AFTER`.
            Not used for other query, key, value modes.
        """

        if n_query_heads < n_key_value_heads or n_query_heads % n_key_value_heads != 0:
            raise ValueError(
                f"Number of query heads ({n_query_heads}) must be a multiple of key/value heads ({n_key_value_heads})"
            )

        return cls(
            n_query_heads=n_query_heads,
            n_key_value_heads=n_key_value_heads,
            qkv_split=qkv_split,
        )


class QkvMode(Enum):
    """
    How the query, key and value projections are handled in
    the self-attention layer.
    """

    #: ``SEPARATE`` - Use separate projections for query, key and value.
    SEPARATE = 0

    #: ``MERGED_SPLIT_BEFORE`` - Use a merged projection for query,
    #: key and value, and split heads before splitting the query, key
    #: and value representations. This ordering is incompatible with
    #: head sharing in keys or values.
    MERGED_SPLIT_BEFORE = 1

    #: ``MERGED_SPLIT_AFTER`` - Use a merged projection for query,
    #: key and value, and split heads after splitting the query, key
    #: and value representations.
    MERGED_SPLIT_AFTER = 2


class AttentionLinearBiases(Module):
    """
    ALiBi: Linear biases for attention (`Press et al., 2022`_).

    .. _Press et al., 2022: https://arxiv.org/abs/2108.12409
    """

    slopes: Tensor

    def __init__(
        self,
        *,
        n_attention_heads: int,
        is_causal: bool,
        is_inverted: bool,
    ) -> None:
        """
        Construct an ALiBi module.

        :param n_attention_heads:
            Number of attention heads.
        :param is_causal:
            Use causal attention.
        :param invert:
            If ``True``, the biases are inverted, i.e.,
            penalties become rewards.
        """
        super().__init__()

        self.is_causal = is_causal
        self.invert = is_inverted
        slopes = self._calculate_slopes(n_attention_heads)
        self.register_buffer("slopes", slopes, persistent=False)

    def _calculate_slopes(self, n_attention_heads: int) -> Tensor:
        """
        Calculate the linear bias slopes for a given number
        of attention heads.

        :param n_attention_heads:
            Number of attention heads.
        :returns:
            Head slope tensor.

            *Shape:* ``(1, heads, 1, 1)``

        :meta private:
        """

        def _slopes_with_step(n_attention_heads, *, step=1):
            ratio = 2.0 ** (-8.0 / n_attention_heads)
            return ratio ** torch.arange(1, 1 + n_attention_heads, step)

        # The slope as proposed in the ALiBi paper would be:
        #
        # return _slopes_with_step(n_attention_heads)
        #
        # However the authors note in their implementation that using powers
        # of 2 for n in the ratio 2**(-8/n) of the geometric sequence for
        # slopes has better properties.
        #
        # Most implementations use powers of two in the following
        # manner: if the number of heads is not a power of 2, then we find
        # k=the largest power of 2 in 1..n. The slopes are then computed
        # as the concatenation of:
        #
        # - The slopes for 1..k.
        # - The slopes for 1..2*k with step 2, taking the first n-k elements.

        # k is the largest power of 2 in 1..n.
        k = 1 << ((n_attention_heads).bit_length() - 1)
        slopes = _slopes_with_step(k)

        if n_attention_heads != k:
            remaining_heads = n_attention_heads - k
            slopes_rest = _slopes_with_step(2 * k, step=2)[:remaining_heads]
            slopes = torch.cat([slopes, slopes_rest])

        return slopes.view(1, -1, 1, 1)

    def calculate_biases(self, seq_len: int) -> Tensor:
        """
        Calculate the linear bias tensor upto a given (key) sequence length.

        :param seq_len:
            Maximum number of timesteps to calculate.
        :returns:
            Multi-headed linear bias tensor.

            *Shape:* ``(1, heads, seq_len, seq_len)`` (non-causal) or
            ``(1, heads, 1, seq_len)`` (causal)

        :meta private:
        """
        if self.is_causal:
            distances = torch.arange(1 - seq_len, 1)
        else:
            distances = torch.arange(seq_len) - torch.arange(seq_len).view(-1, 1)
            distances = distances.abs().mul(-1).view(1, 1, seq_len, seq_len)

        if self.invert:
            distances += seq_len - 1

        return distances.to(self.slopes.device) * self.slopes

    def forward(self, *, attention_scores: Tensor, inplace: bool = True) -> Tensor:
        """
        Apply linear biases to (unmasked) attention scores.

        :param attention_scores:
            Attention scores.

            *Shape:* ``(batch_size, heads, query_len, key_len)``
        :param inplace:
            Update attention scores inplace.
        :returns:
            Attention scores with linear biases.

            *Shape:* ``(batch_size, heads, query_len, key_len)``
        """
        if not inplace:
            attention_scores = attention_scores.clone()

        biases = self.calculate_biases(attention_scores.size(-1)).to(
            dtype=attention_scores.dtype, device=attention_scores.device
        )
        return attention_scores + biases


class AttentionScorer(Module, ABC):
    """
    Base class of attention scoring implementations.
    """

    @abstractmethod
    def forward(
        self,
        *,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        attention_mask: AttentionMask,
        use_causal_mask: bool,
    ) -> Tensor:
        """
        Apply attention scores to the given key, query and value.

        Sequence elements that are marked with ``False`` in the attention mask
        are ignored by the attention mechanism (if a mask is provided).

        :param query:
            Query tensor.

            *Shape:* ``(batch_size, heads, seq_len, width)``
        :param key:
            Key tensor.

            *Shape:* ``(batch_size, heads, seq_len, width)``
        :param value:
            Value tensor.

            *Shape:* ``(batch_size, heads, seq_len, width)``
        :param attention_mask:

            Attention mask. Sequence elements for which the corresponding mask
            element is set to ``False`` are ignored in attention.
        :param use_causal_mask:
            Mask out succeeding sequence elements when ``True``.
        :returns:
            Attention values.

            *Shape:* ``(batch_size, heads, seq_len, width)``
        """
        ...


class ScaledDotProductAttention(AttentionScorer):
    """
    Scaled dot-product attention (`Vaswani et al., 2017`_).

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    linear_biases: Optional[AttentionLinearBiases]

    def __init__(
        self, *, dropout_prob: float, linear_biases: Optional[AttentionLinearBiases]
    ):
        """
        Construct a scaled dot-product attention module.

        :param dropout_prob:
            Dropout to apply to the final hidden representation.
        :param linear_biases:
            ALiBi (`Press et al., 2022`_) for attention scores.
            Not applied if ``None``.

        .. _Press et al., 2022: https://arxiv.org/abs/2108.12409
        """
        super().__init__()

        self.dropout = Dropout(p=dropout_prob)
        self.linear_biases = linear_biases

    def forward(
        self,
        *,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        attention_mask: AttentionMask,
        use_causal_mask: bool,
    ) -> Tensor:
        combined_mask = attention_mask
        if use_causal_mask:
            causal_mask = create_causal_mask(query, key)
            combined_mask = combined_mask.merge_mask(causal_mask)

        if _TORCH_SDP.get():
            logit_mask = combined_mask.logit_mask(query.dtype)

            # Add AliBi to the logit mask
            if self.linear_biases is not None:
                biases = self.linear_biases.calculate_biases(key.size(-2)).to(
                    dtype=query.dtype, device=query.device
                )
                bool_mask = combined_mask.bool_mask
                logit_mask = torch.where(bool_mask, biases, logit_mask)

            # We can't pass a bool mask, because it is currently broken:
            # https://github.com/pytorch/pytorch/issues/103749
            attn_values = F.scaled_dot_product_attention(
                query=query,
                key=key,
                value=value,
                attn_mask=logit_mask,
                dropout_p=self.dropout_prob if self.training else 0.0,
            )

            # Torch SDP returns NaNs for pieces where every is piece masked out.
            # These errors propagate because zero attention times NaN is NaN.
            # Since the representations of these pieces don't matter anyway, we
            # will just zero them out.
            #
            # One issue is that values have shape
            #
            # [batch_len, n_heads, key_len, hidden_size]
            #
            # whereas masks have the shape
            #
            # [batch_len, 1, query_len, key_len]
            #
            # So we can only do this when we have attention masks where
            # the query length it not specified, which are typically 'pure'
            # attention masks (not causal maskes or combined masks):
            #
            # [batch_len, 1, 1, key_len]
            #
            # Doing this properly requires a redesign of our AttentionMask
            # class.
            assert (
                attention_mask.bool_mask.size(-2) == 1
            ), "Torch SDP does not support attention masks with non-broadcastable query length yet"
            return torch.where(
                attention_mask.bool_mask.transpose(-1, -2), attn_values, 0.0
            )
        else:
            width = key.shape[-1]
            attn_scores = query @ key.transpose(-2, -1)
            attn_scores /= math.sqrt(width)

            if self.linear_biases is not None:
                attn_scores = self.linear_biases(attention_scores=attn_scores)

            attn_scores = combined_mask.apply_logit_mask(attn_scores)
            attn_weights = attn_scores.softmax(dim=-1)
            attn_values = self.dropout(attn_weights @ value)

            return attn_values


class SelfAttention(Module):
    """
    Transformer self-attention layer (`Vaswani et al., 2017`_).

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    rotary_embeds: Optional[QueryKeyRotaryEmbeddings]

    def __init__(
        self,
        *,
        attention_heads: AttentionHeads,
        attention_scorer: AttentionScorer,
        hidden_width: int,
        qkv_mode: QkvMode,
        rotary_embeds: Optional[QueryKeyRotaryEmbeddings] = None,
        use_bias: bool,
        device: Optional[torch.device] = None,
    ):
        """
        Construct a self-attention layer with rotary position embeddings
        and attention linear biases.

        :param attention_heads:
            Attention head configuration.
        :param attention_scorer:
            Attention scorer used to calculate the attention values.
        :param hidden_width:
            Hidden width of the layer.
        :param qkv_mode:
            Handling mode for query, key and value.
        :param rotary_embeds:
            Rotary embeddings. Rotary embeddings will not be used when set
            to ``None``.
        :param use_bias:
            Use biases for linear layers.
        :param device:
            Device on which the module is to be initialized.
        """

        super().__init__()

        self.attention_heads = attention_heads
        if hidden_width % attention_heads._n_query_heads != 0:
            raise ValueError(
                f"The hidden width of the transformer ({hidden_width}) must be "
                f"divisible by the number of query heads ({attention_heads._n_query_heads})"
            )

        self.head_width = hidden_width // attention_heads._n_query_heads
        self.qkv_mode = qkv_mode

        self.rotary_embeds = rotary_embeds

        self.attention_scorer = attention_scorer

        if (
            qkv_mode == QkvMode.MERGED_SPLIT_BEFORE
            and attention_heads._n_query_heads != attention_heads._n_key_value_heads
        ):
            raise ValueError(
                "QkvMode.MERGED_SPLIT_BEFORE is incompatible with key/value head sharing"
            )

        key_value_width = attention_heads._n_key_value_heads * self.head_width
        if qkv_mode == QkvMode.SEPARATE:
            self.query = Linear(
                hidden_width,
                hidden_width,
                bias=use_bias,
                device=device,
            )
            self.key = Linear(
                hidden_width,
                key_value_width,
                bias=use_bias,
                device=device,
            )
            self.value = Linear(
                hidden_width,
                key_value_width,
                bias=use_bias,
                device=device,
            )
        else:
            self.input = Linear(
                hidden_width,
                hidden_width + 2 * key_value_width,
                bias=use_bias,
                device=device,
            )

        self.output = Linear(hidden_width, hidden_width, bias=use_bias, device=device)

    def forward(
        self,
        input: Tensor,
        attention_mask: AttentionMask,
        use_causal_mask: bool = False,
        cache: Optional[KeyValueCache] = None,
        store_cache: bool = False,
        positions: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Optional[KeyValueCache]]:
        """
        Apply self-attention layer to the input.

        :param input:
            Input to apply self-attention to.

            *Shape:* ``(batch_size, seq_len, width)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            in attention.
        :param use_causal_mask:
            Mask out succeeding sequence elements when ``True``.
        :param cache:
            Key/value cache to avoid recomputing key/value representations
            for tokens that were previously seen.
        :param store_cache:
            Whether to cache the key/value representations for future reuse.
        :param positions:
            Input positions. Positions are needed to look up rotary embeddings.
            Normally, these positions are calculated automatically. But if the
            positions deviate for some reason, they can be provided through this
            argument.

            *Shape:* ``(batch_size, seq_len)``
        :returns:
            Layer output.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        query, key, value = self._query_key_value(input)

        if self.rotary_embeds is not None:
            query, key = self.rotary_embeds(
                query=query, key=key, cache=cache, positions=positions
            )

        if cache is not None:
            cache_k = cache.key
            cache_v = cache.value

            key = torch.cat([cache_k, key], dim=-2)
            value = torch.cat([cache_v, value], dim=-2)

        attn = self.attention_scorer(
            query=query,
            key=key,
            value=value,
            attention_mask=attention_mask,
            use_causal_mask=use_causal_mask,
        )

        attn = combine_heads(attn)

        output = self.output(attn)

        if store_cache:
            return output, KeyValueCache(key=key, value=value)
        else:
            return output, None

    def _query_key_value(self, input: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Compute query, key and value representations for the input.

        :param input:
            Input

            *Shape:* ``(batch_size, seq_len, hidden_width)``
        :returns:
            Query, key, value

            *Shape:* ``(batch_size, head, seq_len, width_per_head)``
        """

        n_key_value_heads = self.attention_heads._n_key_value_heads
        n_query_heads = self.attention_heads._n_query_heads
        if self.qkv_mode == QkvMode.SEPARATE:
            query, key, value = self._query_key_value_separate(input)
        elif self.qkv_mode == QkvMode.MERGED_SPLIT_BEFORE:
            query, key, value = self._query_key_value_merged_split_before(input)
        else:
            query, key, value = self._query_key_value_merged_split_after(input)

        if n_key_value_heads != 1 and n_key_value_heads != n_query_heads:
            # If all the key/value representations are shared, their head
            # dimension is 1 and the heads will broadcast in later ops.
            # However, we have multiple heads, we have to broadcast the
            # heads explicitly. Like splitting, there are multiple ways
            # to do this. The current interleaving repeat is used for
            # compatibility with falcon.
            n_tiling = n_query_heads // n_key_value_heads
            key = key.repeat_interleave(n_tiling, dim=1)
            value = value.repeat_interleave(n_tiling, dim=1)

        return query, key, value

    def _query_key_value_separate(self, input: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Compute query, key and value representations for the input, using
        separate query, key, and value projections.

        :param input:
            Input

            *Shape:* ``(batch_size, seq_len, hidden_width)``
        :returns:
            Query, key, value

            *Shape:* ``(batch_size, n_head, seq_len, width_per_head)``
        """
        query = self.query(input)
        key = self.key(input)
        value = self.value(input)

        query = split_heads(query, self.attention_heads._n_query_heads)
        key = split_heads(key, self.attention_heads._n_key_value_heads)
        value = split_heads(value, self.attention_heads._n_key_value_heads)

        return query, key, value

    def _query_key_value_merged_split_before(
        self, input: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Compute query, key and value representations for the input, using
        merged query, key, and value projections. Split heads before
        splitting query, key, and value.

        :param input:
            Input

            *Shape:* ``(batch_size, seq_len, hidden_width)``
        :returns:
            Query, key, value

            *Shape:* ``(batch_size, n_head, seq_len, width_per_head)``
        """
        proj = self.input(input)
        # Same number of heads for query, key and value
        # since we cannot share heads in this mode.
        proj = split_heads(proj, self.attention_heads._n_query_heads)
        query, key, value = proj.chunk(3, dim=-1)

        return query, key, value

    def _query_key_value_merged_split_after(
        self, input: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Compute query, key and value representations for the input, using
        merged query, key, and value projections. Split heads after
        splitting query, key, and value.

        :param input:
            Input

            *Shape:* ``(batch_size, seq_len, hidden_width)``
        :returns:
            Query, key, value

            *Shape:* ``(batch_size, n_head, seq_len, width_per_head)``
        """
        proj = self.input(input)

        # This splitting method is used by Falcon. There are other
        # (simpler) splitting methods. For instance, we could group all
        # the heads together and then split them. In the future we
        # could split conditional on a field in AttentionHeads.
        query, key, value = self.attention_heads._qkv_split.split(
            projection=proj,
            head_width=self.head_width,
            n_query_heads=self.attention_heads._n_query_heads,
            n_key_value_heads=self.attention_heads._n_key_value_heads,
        )

        return query, key, value


def split_heads(input: Tensor, n_heads: int) -> Tensor:
    """
    Split the input by attention head. The caller must validate
    that the innermost dimension is divisable by the number of
    heads.

    :param input:
        Tensor to split by head.

        *Shape:* ``(batch_size, seq_len, hidden_width)``
    :param n_heads:
        Number of attention heads.
    :returns:
        Tensor spilt by head.

        *Shape:* ``(batch_size, head, seq_len, width_per_head)``
    """
    batch_size, seq_len, model_width = input.size()
    assert model_width % n_heads == 0
    head_width = model_width // n_heads

    return input.view(batch_size, seq_len, n_heads, head_width).transpose(1, 2)


def combine_heads(input: Tensor) -> Tensor:
    """
    Combine the split attention head representations.

    :param input:
        Tensor split by head.

        *Shape:* ``(batch_size, head, seq_len, width_per_head)``
    :returns:
        Merged tensor.

        *Shape:* ``(batch_size, seq_len, hidden_width)``
    """
    batch_size, head, seq_len, model_width = input.size()
    return (
        input.transpose(1, 2).contiguous().view(batch_size, seq_len, head * model_width)
    )



from typing import Any, Dict, List, Mapping, Optional, Tuple

import torch

from ...util.string import StringTransform, StringTransformations
from ..hf_hub.conversion import (
    CommonHFKeys,
    HFConfigKey,
    HFConfigKeyDefault,
    HFSpecificConfig,
    config_from_hf,
    config_to_hf,
)
from .config import ALBERTConfig

# Order-dependent.
HF_PARAM_KEY_TRANSFORMS: List[StringTransform] = [
    # Prefixes.
    StringTransformations.remove_prefix("albert.", reversible=False),
    StringTransformations.regex_sub(
        (r"^encoder\.(embedding_|albert_layer)", "\\1"),
        (r"^(embedding_|albert_layer)", "encoder.\\1"),
    ),
    # Layer groups
    StringTransformations.regex_sub(
        (r"^albert_layer_groups\.", "groups."), (r"^groups\.", "albert_layer_groups.")
    ),
    # Inner layers.
    StringTransformations.sub(".albert_layers.", ".group_layers."),
    # Attention blocks.
    StringTransformations.sub(".attention.", ".mha."),
    StringTransformations.sub(".mha.LayerNorm", ".attn_residual_layer_norm"),
    StringTransformations.sub(".mha.dense", ".mha.output"),
    # Pointwise feed-forward layers.
    StringTransformations.sub(".ffn.", ".ffn.intermediate."),
    StringTransformations.sub(".ffn_output.", ".ffn.output."),
    StringTransformations.sub(".full_layer_layer_norm.", ".ffn_residual_layer_norm."),
    # Embeddings.
    StringTransformations.replace(
        "embeddings.word_embeddings.weight", "embeddings.piece_embeddings.weight"
    ),
    StringTransformations.replace(
        "embeddings.token_type_embeddings.weight", "embeddings.type_embeddings.weight"
    ),
    StringTransformations.replace(
        "embeddings.position_embeddings.weight", "embeddings.position_embeddings.weight"
    ),
    StringTransformations.replace(
        "embeddings.LayerNorm.weight", "embeddings.embed_output_layer_norm.weight"
    ),
    StringTransformations.replace(
        "embeddings.LayerNorm.bias", "embeddings.embed_output_layer_norm.bias"
    ),
    # Embedding projection.
    StringTransformations.replace(
        "embedding_hidden_mapping_in.weight", "embeddings.projection.weight"
    ),
    StringTransformations.replace(
        "embedding_hidden_mapping_in.bias", "embeddings.projection.bias"
    ),
]


class HFConfigKeys:
    @staticmethod
    def conv_n_layers_per_group(config: ALBERTConfig) -> int:
        return config.layer.n_layers_per_group

    @staticmethod
    def conv_n_hidden_groups(config: ALBERTConfig) -> int:
        return config.layer.n_hidden_groups

    INNER_GROUP_NUM = HFConfigKey(
        "inner_group_num",
        "n_layers_per_group",
        lambda c: HFConfigKeys.conv_n_layers_per_group(c),
    )
    NUM_HIDDEN_GROUPS = HFConfigKey(
        "num_hidden_groups",
        "n_hidden_groups",
        lambda c: HFConfigKeys.conv_n_hidden_groups(c),
    )


HF_CONFIG_KEYS: List[Tuple[HFConfigKey, Optional[HFConfigKeyDefault]]] = [
    (CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB, None),
    (CommonHFKeys.DTYPE, HFConfigKeyDefault("float32")),
    (CommonHFKeys.EMBEDDING_SIZE, None),
    (CommonHFKeys.HIDDEN_DROPOUT_PROB, None),
    (CommonHFKeys.HIDDEN_SIZE, None),
    (CommonHFKeys.HIDDEN_ACT, None),
    (CommonHFKeys.INTERMEDIATE_SIZE, None),
    (CommonHFKeys.LAYER_NORM_EPS, None),
    (CommonHFKeys.NUM_ATTENTION_HEADS_UNIFORM, None),
    (CommonHFKeys.NUM_HIDDEN_LAYERS, None),
    (CommonHFKeys.VOCAB_SIZE, None),
    (CommonHFKeys.TYPE_VOCAB_SIZE, None),
    (CommonHFKeys.MAX_POSITION_EMBEDDINGS, None),
    (HFConfigKeys.INNER_GROUP_NUM, None),
    (HFConfigKeys.NUM_HIDDEN_GROUPS, None),
]

HF_SPECIFIC_CONFIG = HFSpecificConfig(
    architectures=["AlbertModel"], model_type="albert"
)


def _config_from_hf(hf_config: Mapping[str, Any]) -> ALBERTConfig:
    kwargs = config_from_hf("ALBERT", hf_config, HF_CONFIG_KEYS)
    return ALBERTConfig(
        model_max_length=CommonHFKeys.MAX_POSITION_EMBEDDINGS.get_kwarg(kwargs),
        **kwargs
    )


def _config_to_hf(curated_config: ALBERTConfig) -> Dict[str, Any]:
    out = config_to_hf(curated_config, [k for k, _ in HF_CONFIG_KEYS])
    return HF_SPECIFIC_CONFIG.merge(out)



from ..layers.feedforward import PointwiseFeedForward
from .activations import Activation, GELUFast, GELUNew
from .attention import (
    AttentionHeads,
    AttentionLinearBiases,
    AttentionMask,
    AttentionScorer,
    QkvMode,
    QkvSplit,
    ScaledDotProductAttention,
    SelfAttention,
    enable_torch_sdp,
)
from .cache import CacheProtocol, KeyValueCache
from .embeddings import (
    QueryKeyRotaryEmbeddings,
    RotaryEmbeddings,
    SinusoidalPositionalEmbedding,
)
from .normalization import RMSNorm
from .transformer import (
    DecoderLayer,
    EmbeddingDropouts,
    EmbeddingLayerNorms,
    EncoderLayer,
    TransformerDropouts,
    TransformerEmbeddings,
    TransformerLayerNorms,
)

__all__ = [
    "Activation",
    "AttentionHeads",
    "AttentionLinearBiases",
    "AttentionMask",
    "AttentionScorer",
    "CacheProtocol",
    "DecoderLayer",
    "EmbeddingDropouts",
    "EmbeddingLayerNorms",
    "EncoderLayer",
    "GELUFast",
    "GELUNew",
    "KeyValueCache",
    "PointwiseFeedForward",
    "QkvMode",
    "QkvSplit",
    "QueryKeyRotaryEmbeddings",
    "RMSNorm",
    "RotaryEmbeddings",
    "ScaledDotProductAttention",
    "SelfAttention",
    "SinusoidalPositionalEmbedding",
    "TransformerDropouts",
    "TransformerEmbeddings",
    "TransformerLayerNorms",
    "enable_torch_sdp",
]



from typing import Iterable, Tuple


def remove_pieces_from_sequence(
    input: Iterable[int], pieces_to_remove: Tuple[int, ...]
) -> Iterable[int]:
    return (piece for piece in input if piece not in pieces_to_remove)



from dataclasses import dataclass

import torch

from ..bert import BERTConfig


@dataclass
class RoBERTaConfig(BERTConfig):
    """
    RoBERTa (`Liu et al., 2019`_) model configuration.

    .. _Liu et al., 2019: https://arxiv.org/abs/1907.11692
    """

    def __init__(
        self,
        *args,
        dtype: torch.dtype = torch.float32,
        layer_norm_eps=1e-05,
        n_positions=514,
        padding_id=1,
        n_types=1,
        n_pieces=50265,
        **kwargs
    ):
        """
        :param dtype:
            Data type to use for model parameters.
        :param embedding_width:
            Width of the embedding representations.
        :param hidden_width:
            Width of the transformer hidden layers.
        :param intermediate_width:
            Width of the intermediate projection layer in the
            point-wise feed-forward layer.
        :param n_attention_heads:
            Number of self-attention heads.
        :param n_hidden_layers:
            Number of hidden layers.
        :param attention_probs_dropout_prob:
            Dropout probabilty of the self-attention layers.
        :param hidden_dropout_prob:
            Dropout probabilty of the point-wise feed-forward and
            embedding layers.
        :param activation:
            Activation used by the pointwise feed-forward layers.
        :param n_pieces:
            Size of main vocabulary.
        :param n_types:
            Size of token type vocabulary.
        :param n_positions:
            Maximum length of position embeddings.
        :param model_max_length:
            Maximum length of model inputs.
        :param layer_norm_eps:
            Epsilon for layer normalization.
        :param padding_id:
            Index of the padding meta-token.
        """
        super(RoBERTaConfig, self).__init__(
            *args,
            layer_norm_eps=layer_norm_eps,
            n_positions=n_positions,
            n_types=n_types,
            n_pieces=n_pieces,
            **kwargs
        )
        self.dtype = dtype
        self.padding_id = padding_id



from enum import Enum
from typing import TYPE_CHECKING, Callable, Iterable, Mapping

import torch

from ..._compat import has_safetensors
from ...repository.file import RepositoryFile

if TYPE_CHECKING:
    import safetensors


class ModelCheckpointType(Enum):
    """
    Types of model checkpoints supported by Curated Transformers.
    """

    #: PyTorch `checkpoint<https://pytorch.org/docs/stable/generated/torch.save.html>`_.
    PYTORCH_STATE_DICT = 0

    #: Hugging Face `Safetensors <https://github.com/huggingface/safetensors>`_ checkpoint.
    SAFE_TENSORS = 1

    @property
    def loader(
        self,
    ) -> Callable[[Iterable[RepositoryFile]], Iterable[Mapping[str, torch.Tensor]]]:
        checkpoint_type_to_loader = {
            ModelCheckpointType.PYTORCH_STATE_DICT: _load_pytorch_state_dicts_from_checkpoints,
            ModelCheckpointType.SAFE_TENSORS: _load_safetensor_state_dicts_from_checkpoints,
        }
        return checkpoint_type_to_loader[self]

    @property
    def pretty_name(self) -> str:
        if self == ModelCheckpointType.PYTORCH_STATE_DICT:
            return "PyTorch StateDict"
        elif self == ModelCheckpointType.SAFE_TENSORS:
            return "SafeTensors"
        else:
            return ""


def _load_safetensor_state_dicts_from_checkpoints(
    checkpoints: Iterable[RepositoryFile],
) -> Iterable[Mapping[str, torch.Tensor]]:
    if not has_safetensors:
        raise ValueError(
            "The `safetensors` library is required to load models from Safetensors checkpoints"
        )

    import safetensors.torch

    for checkpoint in checkpoints:
        # Prefer to load from a path when possible. Since loading from a file
        # temporarily puts the checkpoint in memory twice.
        if checkpoint.path is not None:
            # Map to CPU first to support all devices.
            state_dict = safetensors.torch.load_file(checkpoint.path, device="cpu")
        else:
            with checkpoint.open() as f:
                # This has memory overhead, since Safetensors does not have
                # support for loading from a file object and cannot use
                # the bytes in-place.
                checkpoint_bytes = f.read()
                state_dict = safetensors.torch.load(checkpoint_bytes)
        yield state_dict


def _load_pytorch_state_dicts_from_checkpoints(
    checkpoints: Iterable[RepositoryFile],
) -> Iterable[Mapping[str, torch.Tensor]]:
    for checkpoint in checkpoints:
        with checkpoint.open() as f:
            # Map to CPU first to support all devices.
            state_dict = torch.load(
                f, map_location=torch.device("cpu"), weights_only=True
            )
        yield state_dict



from abc import ABC, abstractmethod
from typing import Set


class Quantizable(ABC):
    """
    Mixin class for models that are quantizable.

    A module using this mixin provides the necessary configuration
    and parameter information to quantize it on-the-fly during the
    module loading phase.
    """

    @classmethod
    @abstractmethod
    def modules_to_not_quantize(cls) -> Set[str]:
        """
        Return a set of prefixes that specify which
        modules are to be ignored during quantization.

        :returns:
            Set of module prefixes.

            If empty, all submodules will be quantized.
        """
        raise NotImplementedError



import dataclasses
from dataclasses import dataclass
from typing import List, Optional, Union

ChunkT = Union["SpecialPieceChunk", "TextChunk"]
MergedChunkT = Union["MergedSpecialPieceChunk", "TextChunk"]


@dataclass
class MergedSpecialPieceChunk:
    """
    A chunk that contains a special piece. This piece is not tokenized, but
    looked up directly in the vocabulary.

    :param piece:
        Piece to look up in the vocabulary.
    """

    piece: str


@dataclass
class SpecialPieceChunk:
    """
    A chunk that contains a special piece. This piece is not tokenized, but
    looked up directly in the vocabulary. Can additionally store strings that
    should be appended to a text chunk before or prepended to a text chunk
    after the special piece.

    :param piece:
        Piece to look up in the vocabulary.
    :param after:
        Text to prepend to the succeeding text chunk.
    :param before:
        Text to append to the preceding text chunk.
    """

    piece: str
    after: Optional[str] = None
    before: Optional[str] = None


@dataclass
class TextChunk:
    """
    A chunk of text that should be tokenized.

    :param text:
        Text that should be tokenized.
    """

    text: str


class MergedInputChunks(List[MergedChunkT]):
    """
    A list of chunks in which consecutive text chunks and before/after
    texts of special piece chunks are merged.
    """

    pass


class InputChunks(List[ChunkT]):
    """
    A list of chunks.
    """

    def merge_text_chunks(self) -> MergedInputChunks:
        """
        Merge multiple contiguous text chunks and before/after text
        in special piece chunks.
        """
        new_chunks = MergedInputChunks()
        for chunk in self:
            last_is_text = new_chunks and isinstance(new_chunks[-1], TextChunk)
            if isinstance(chunk, TextChunk):
                if last_is_text:
                    new_chunks[-1].text += chunk.text  # type: ignore[union-attr]
                else:
                    new_chunks.append(dataclasses.replace(chunk))
            elif isinstance(chunk, SpecialPieceChunk):
                if chunk.before:
                    if last_is_text:
                        new_chunks[-1].text += chunk.before  # type: ignore[union-attr]
                    else:
                        new_chunks.append(TextChunk(chunk.before))
                new_chunks.append(MergedSpecialPieceChunk(chunk.piece))
                if chunk.after:
                    new_chunks.append(TextChunk(chunk.after))
            else:
                raise ValueError(f"Unknown chunk type: {type(chunk).__name__}")

        return new_chunks



from typing import List, TypeVar

from ..models.gpt_neox.causal_lm import GPTNeoXCausalLM
from ..tokenizers.chunks import InputChunks, SpecialPieceChunk, TextChunk
from ..tokenizers.tokenizer import Tokenizer
from .config import SampleGeneratorConfig
from .default_generator import DefaultGenerator

INSTRUCTION_KEY = "### Instruction:"
RESPONSE_KEY = "### Response:"
END_KEY = "### End"
INTRO_BLURB = "Below is an instruction that describes a task. Write a response that appropriately completes the request."


# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="DollyV2Generator")


class DollyV2Generator(DefaultGenerator):
    """
    Generator for Dolly v2 model variants.
    """

    def __init__(self, tokenizer: Tokenizer, causal_lm: GPTNeoXCausalLM):
        """
        Construct a Dolly v2 generator.

        :param tokenizer:
            A Dolly v2 tokenizer.
        :param causal_lm:
            A Dolly v2 causal language model.

        """
        eos_id = tokenizer.tokenizer.token_to_id(END_KEY)
        super().__init__(
            tokenizer,
            causal_lm,
            default_config=SampleGeneratorConfig(
                eos_id=eos_id,
                max_generated_pieces=256,
                top_p=0.92,
            ),
        )

    def preprocess_prompts(self, prompts: List[str]) -> List[InputChunks]:
        return [
            InputChunks(
                [
                    TextChunk(INTRO_BLURB),
                    # Dolly is really picky about getting the correct number of line
                    # breaks, so the double line breaks in this chunk and the last
                    # chunk are intentional.
                    SpecialPieceChunk(INSTRUCTION_KEY, before="\n\n", after="\n"),
                    TextChunk(prompt),
                    SpecialPieceChunk(RESPONSE_KEY, before="\n\n", after="\n"),
                ]
            )
            for prompt in prompts
        ]



import pytest
from huggingface_hub import HfFileSystem

from curated_transformers.repository.fsspec import FsspecArgs
from curated_transformers.tokenizers import AutoTokenizer

_MODELS = [
    # tokenizer.json-only
    ("EleutherAI/gpt-neox-20b", "9369f145ca7b66ef62760f9351af951b2d53b77f"),
    # sentencepiece binary
    ("openlm-research/open_llama_13b", "b6d7fde8392250730d24cc2fcfa3b7e5f9a03ce8"),
    # tokenizer config, but no class name
    ("GroNLP/bert-base-dutch-cased", "484ff5cec2ad42b434537dadd901d9b8e2b64cd2"),
]


@pytest.mark.parametrize("model_revision", _MODELS)
def test_auto_tokenizer(model_revision):
    name, revision = model_revision
    AutoTokenizer.from_hf_hub(name=name, revision=revision)


@pytest.mark.slow
@pytest.mark.parametrize("model_revision", _MODELS)
def test_auto_tokenizer_fsspec(model_revision):
    name, revision = model_revision
    AutoTokenizer.from_fsspec(
        fs=HfFileSystem(), model_path=name, fsspec_args=FsspecArgs(revision=revision)
    )
    AutoTokenizer.from_hf_hub(name=name, revision=revision)


def test_cannot_infer():
    # This repo/revision does not have a tokenizer and doesn't match a
    # legacy tokenizer.
    with pytest.raises(ValueError, match=r"Cannot infer tokenizer for repo"):
        AutoTokenizer.from_hf_hub(
            name="explosion-testing/falcon-test",
            revision="235f4b64c489e33ae7e40163ab1f266bbe355651",
        )



import warnings
from dataclasses import dataclass
from typing import IO, Any, Dict, Optional

from fsspec import AbstractFileSystem

from .repository import Repository, RepositoryFile
from .transaction import TransactionContext


@dataclass
class FsspecArgs:
    """
    Convenience wrapper for additional fsspec arguments.
    """

    kwargs: Dict[str, Any]

    def __init__(self, **kwargs):
        """
        Keyword arguments are passed through to the fsspec implementation.
        Construction may raise in the future when reserved arguments like
        ``mode`` or ``encoding`` are used.
        """
        # Future improvement: raise on args that are used by fsspec methods,
        # e.g. `mode` or `encoding`.
        self.kwargs = kwargs


class FsspecFile(RepositoryFile):
    """
    Repository file on an `fsspec`_ filesystem.

    .. _fsspec: https://filesystem-spec.readthedocs.io/en/latest/
    """

    def __init__(
        self,
        fs: AbstractFileSystem,
        path: str,
        fsspec_args: Optional[FsspecArgs] = None,
    ):
        """
        Construct an fsspec file representation.

        :param fs:
            The filesystem.
        :param path:
            The path of the file on the filesystem.
        :param fsspec_args:
            Implementation-specific arguments to pass to fsspec filesystem
            operations.
        """
        super().__init__()
        self._fs = fs
        self._path = path
        self._fsspec_args = FsspecArgs() if fsspec_args is None else fsspec_args

    def open(self, mode: str = "rb", encoding: Optional[str] = None) -> IO:
        if ("r" in mode or "+" in mode) and not self.exists():
            raise FileNotFoundError(
                f"Cannot find fsspec path {self._fs.unstrip_protocol(self._path)}"
            )

        try:
            return self._fs.open(
                self._path, mode=mode, encoding=encoding, **self._fsspec_args.kwargs
            )
        except Exception as e:
            raise OSError(
                f"Cannot open fsspec path {self._fs.unstrip_protocol(self._path)}"
            ) from e

    @property
    def path(self) -> Optional[str]:
        return None

    def exists(self) -> bool:
        return self._fs.exists(self._path, **self._fsspec_args.kwargs)


class FsspecRepository(Repository):
    """
    Repository using a filesystem that uses the `fsspec`_ interface.

    .. _fsspec: https://filesystem-spec.readthedocs.io/en/latest/

    :param fs:
        The filesystem.
    :param path:
        The the path of the repository within the filesystem.
    :param fsspec_args:
        Additional arguments that should be passed to the fsspec
        implementation.
    """

    def __init__(
        self,
        fs: AbstractFileSystem,
        path: str,
        fsspec_args: Optional[FsspecArgs] = None,
    ):
        super().__init__()
        self.fs = fs
        self.repo_path = path
        self.fsspec_args = FsspecArgs() if fsspec_args is None else fsspec_args

    def file(self, path: str) -> RepositoryFile:
        full_path = f"{self.repo_path}/{path}"
        return FsspecFile(self.fs, full_path, self.fsspec_args)

    def pretty_path(self, path: Optional[str] = None) -> str:
        if not path:
            return self._protocol
        return f"{self._protocol}/{path}"

    def transaction(self) -> TransactionContext:
        return FsspecTransactionContext(self)

    @property
    def _protocol(self) -> str:
        return self.fs.unstrip_protocol(self.repo_path)


class FsspecTransactionContext(TransactionContext):
    """
    `fsspec`_ transaction context manager.
    TODO: Implement/currently a noop

    .. _fsspec: https://filesystem-spec.readthedocs.io/en/latest/
    """

    def __init__(self, repo: FsspecRepository):
        """
        :param repo:
            The parent repository.
        """
        super().__init__()
        self._repo = repo

    def open(self, path: str, mode: str, encoding: Optional[str] = None) -> IO:
        return self._repo.file(path).open(mode=mode, encoding=encoding)

    @property
    def repo(self) -> Repository:
        return self._repo

    def __enter__(self):
        warnings.warn("Transaction support is currently not implemented for fsspec")
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        pass



import pytest

from curated_transformers.generation import AutoGenerator
from curated_transformers.generation.dolly_v2 import DollyV2Generator
from curated_transformers.generation.falcon import FalconGenerator
from curated_transformers.generation.llama import LlamaGenerator
from curated_transformers.generation.mpt import MPTGenerator


@pytest.mark.slow
def test_auto_generator():
    model_causallm_map = {
        "databricks/dolly-v2-3b": DollyV2Generator,
        "tiiuae/falcon-7b": FalconGenerator,
        "openlm-research/open_llama_3b": LlamaGenerator,
        "mosaicml/mpt-7b": MPTGenerator,
    }

    for name, generator_cls in model_causallm_map.items():
        generator = AutoGenerator.from_hf_hub(name=name)
        assert isinstance(generator, generator_cls)

    with pytest.raises(ValueError, match="Unsupported generator"):
        AutoGenerator.from_hf_hub(name="trl-internal-testing/tiny-random-GPT2Model")



from dataclasses import dataclass
from enum import Enum
from typing import Union

import torch


class Dtype4Bit(str, Enum):
    """
    Data type to use for 4-bit quantization.
    """

    #: ``FP4`` - Float 4-bit.
    FP4 = "fp4"

    #: ``NF4`` - NormalFloat 4-bit.
    NF4 = "nf4"


@dataclass
class _4BitConfig:
    """
    Configuration for ``fp4``/``nf4`` quantization.
    """

    quantization_dtype: Dtype4Bit
    compute_dtype: torch.dtype
    double_quantization: bool


@dataclass
class _8BitConfig:
    """
    Configuration for ``int8`` quantization.
    """

    finetunable: bool
    outlier_threshold: float


@dataclass
class BitsAndBytesConfig:
    """
    Configuration for quantization using the ``bitsandbytes`` library.
    """

    inner: Union[_4BitConfig, _8BitConfig]

    @staticmethod
    def for_8bit(outlier_threshold: float = 6.0, finetunable: bool = False):
        """
        Construct a configuration for ``int8`` quantization.

        :param outlier_threshold:
            Threshold for outlier detection during weight
            decomposition.
        :param finetunable:
            If the quantized model should support fine-tuning after
            quantization.
        """
        return BitsAndBytesConfig(
            _8BitConfig(finetunable=finetunable, outlier_threshold=outlier_threshold)
        )

    @staticmethod
    def for_4bit(
        quantization_dtype: Dtype4Bit = Dtype4Bit.FP4,
        compute_dtype: torch.dtype = torch.bfloat16,
        double_quantization: bool = True,
    ):
        """
        Construct a configuration for ``fp4``/``nf4`` quantization.

        :param quantization_dtype:
            Data type used for storing quantized weights.
        :param compute_dtype:
            Data type used for performing computations.
            Supported types: ``float16``, ``bfloat16``, ``float32``.
        :param double_quantization:
            If the quantization constants should themselves be
            quantized.
        """
        supported_compute_dtypes = (torch.float32, torch.float16, torch.bfloat16)
        if compute_dtype not in supported_compute_dtypes:
            raise ValueError(
                f"Unsupported compute dtype `{compute_dtype}` for quantization, must be one of: {supported_compute_dtypes}"
            )

        return BitsAndBytesConfig(
            _4BitConfig(
                quantization_dtype=quantization_dtype,
                compute_dtype=compute_dtype,
                double_quantization=double_quantization,
            )
        )



from abc import ABC, abstractmethod
from typing import IO, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .repository import Repository


class TransactionContext(ABC):
    """
    A context manager that represents an active transaction in
    a repository.
    """

    @abstractmethod
    def open(self, path: str, mode: str, encoding: Optional[str] = None) -> IO:
        """
        Opens a file as a part of a transaction. Changes to the
        file are deferred until the transaction has completed
        successfully.

        :param path:
            The path to the file on the parent repository.
        :param mode:
            Mode to open the file with (see Python ``open``).
        :param encoding:
            Encoding to use when the file is opened as text.
        :returns:
            An I/O stream.
        :raises FileNotFoundError:
            When the file cannot be found.
        :raises OSError:
            When the file cannot be opened.
        """
        ...

    @property
    @abstractmethod
    def repo(self) -> "Repository":
        """
        :returns:
            The parent repository on which this transaction is performed.
        """
        ...

    @abstractmethod
    def __enter__(self):
        """
        Invoked when the context manager is entered.
        """
        ...

    @abstractmethod
    def __exit__(self, exc_type, exc_value, exc_tb):
        """
        Invoked when the context manager exits.
        """
        ...



import torch
from torch.nn import Identity

try:
    import transformers

    has_hf_transformers = True
except ImportError:
    transformers = None  # type: ignore
    has_hf_transformers = False

has_torch_compile = hasattr(torch, "compile")
if has_torch_compile:
    # torch.compile is not supported on all Python versions and platforms.
    try:
        torch.compile(Identity(), disable=True)
    except:
        has_torch_compile = False

try:
    import safetensors

    has_safetensors = True
except ImportError:
    safetensors = None  # type: ignore
    has_safetensors = False



from pathlib import Path

import pytest
from huggingface_hub import _CACHED_NO_EXIST, try_to_load_from_cache

from curated_transformers.models.bert.encoder import BERTEncoder
from curated_transformers.repository.hf_hub import HfHubRepository
from curated_transformers.repository.repository import ModelRepository
from curated_transformers.util.serde.checkpoint import ModelCheckpointType

from ..compat import has_hf_transformers, has_safetensors
from ..conftest import TORCH_DEVICES
from .util import assert_encoder_output_equals_hf


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_sharded_model_checkpoints(torch_device):
    assert_encoder_output_equals_hf(
        BERTEncoder, "explosion-testing/bert-test-sharded", torch_device
    )


def test_from_hf_hub_to_cache():
    BERTEncoder.from_hf_hub_to_cache(
        name="explosion-testing/bert-test-caching",
        revision="96a29a07d0fa4c24fd2675521add643e3c2581fc",
    )

    expected_checkpoints = [
        "pytorch_model-00001-of-00007.bin",
        "pytorch_model-00002-of-00007.bin",
        "pytorch_model-00003-of-00007.bin",
        "pytorch_model-00004-of-00007.bin",
        "pytorch_model-00005-of-00007.bin",
        "pytorch_model-00006-of-00007.bin",
        "pytorch_model-00007-of-00007.bin",
    ]
    for name in expected_checkpoints:
        assert (
            try_to_load_from_cache(
                repo_id="explosion-testing/bert-test-caching",
                filename=name,
                revision="96a29a07d0fa4c24fd2675521add643e3c2581fc",
            )
            != _CACHED_NO_EXIST
        )


@pytest.mark.skipif(has_safetensors, reason="cannot run with huggingface safetensors")
def test_checkpoint_type_without_safetensors():
    # By default, we expect the torch checkpoint to be loaded
    # even if the safetensor checkpoints are present
    # (as long as the library is not installed).
    ckp_paths, ckp_type = ModelRepository(
        HfHubRepository("explosion-testing/safetensors-test", revision="main")
    ).model_checkpoints()
    assert len(ckp_paths) == 1
    assert ckp_paths[0].path is not None
    assert Path(ckp_paths[0].path).suffix == ".bin"
    assert ckp_type == ModelCheckpointType.PYTORCH_STATE_DICT


@pytest.mark.skipif(not has_safetensors, reason="requires huggingface safetensors")
def test_checkpoint_type_with_safetensors():
    # Since the safetensors library is installed, we should be
    # loading from those checkpoints.
    repo = ModelRepository(
        HfHubRepository("explosion-testing/safetensors-test", revision="main")
    )
    ckp_paths, ckp_type = repo.model_checkpoints()
    assert len(ckp_paths) == 1
    assert Path(ckp_paths[0].path).suffix == ".safetensors"
    assert ckp_type == ModelCheckpointType.SAFE_TENSORS

    encoder = BERTEncoder.from_hf_hub(name="explosion-testing/safetensors-test")


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_fsspec(torch_device):
    assert_encoder_output_equals_hf(
        BERTEncoder,
        "explosion-testing/bert-test",
        torch_device,
        with_fsspec=True,
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_fsspec_sharded(torch_device):
    assert_encoder_output_equals_hf(
        BERTEncoder,
        "explosion-testing/bert-test-sharded",
        torch_device,
        with_fsspec=True,
    )


@pytest.mark.slow
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_fsspec_safetensors(torch_device):
    assert_encoder_output_equals_hf(
        BERTEncoder,
        "explosion-testing/safetensors-test",
        torch_device,
        with_fsspec=True,
    )



from typing import Any, Dict, Mapping, Optional, Type, TypeVar

from curated_tokenizers import SentencePieceProcessor

from ...repository.file import RepositoryFile
from ..hf_hub import LegacyFromHF
from ._fairseq import FAIRSEQ_PIECE_IDS, FairSeqPostEncoder, FairSeqPreDecoder
from .legacy_tokenizer import AddBosEosPreEncoder
from .sentencepiece_tokenizer import SentencePieceTokenizer

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="XLMRTokenizer")


_XLMR_FAIRSEQ_OFFSET = 1


class XLMRPostEncoder(FairSeqPostEncoder):
    def __init__(
        self,
    ):
        """
        Construct a XLM-R post-encoder.
        """
        super(XLMRPostEncoder, self).__init__(
            piece_updater=XLMRPostEncoder._sentencepiece_to_fairseq,
        )

    @staticmethod
    def _sentencepiece_to_fairseq(piece_id: int):
        if piece_id == FAIRSEQ_PIECE_IDS.SPP_UNK:
            return FAIRSEQ_PIECE_IDS.FAIRSEQ_UNK
        elif piece_id == FAIRSEQ_PIECE_IDS.SPP_BOS:
            return FAIRSEQ_PIECE_IDS.FAIRSEQ_BOS
        elif piece_id == FAIRSEQ_PIECE_IDS.SPP_EOS:
            return FAIRSEQ_PIECE_IDS.FAIRSEQ_EOS
        else:
            return piece_id + _XLMR_FAIRSEQ_OFFSET


class XLMRPreDecoder(FairSeqPreDecoder):
    def __init__(
        self,
        *,
        bos_id: int,
        eos_id: int,
    ):
        """
        Construct a XLM-R pre-decoder.

        :param bos_id:
            The piece id used to mark the beginning of a sequence.
        :param eos_id:
            The piece id used to mark the end of a sequence.
        """
        self.bos_id = bos_id
        self.eos_id = eos_id
        super(XLMRPreDecoder, self).__init__(
            bos_id=bos_id,
            eos_id=eos_id,
            piece_updater=XLMRPreDecoder._fairseq_to_sentencepiece,
        )

    @staticmethod
    def _fairseq_to_sentencepiece(piece_id: int):
        if piece_id == FAIRSEQ_PIECE_IDS.FAIRSEQ_UNK:
            return FAIRSEQ_PIECE_IDS.SPP_UNK
        elif piece_id == FAIRSEQ_PIECE_IDS.FAIRSEQ_BOS:
            return FAIRSEQ_PIECE_IDS.SPP_BOS
        elif piece_id == FAIRSEQ_PIECE_IDS.FAIRSEQ_EOS:
            return FAIRSEQ_PIECE_IDS.SPP_EOS
        else:
            return piece_id - _XLMR_FAIRSEQ_OFFSET


class XLMRTokenizer(SentencePieceTokenizer, LegacyFromHF):
    """
    Legacy tokenizer for XLM-RoBERTa (`Conneau et al., 2019`_) models.

    .. _Conneau et al., 2019: https://arxiv.org/abs/1911.02116
    """

    vocab_files: Dict[str, str] = {"model": "sentencepiece.bpe.model"}

    def __init__(
        self,
        *,
        processor: SentencePieceProcessor,
    ):
        """
        Construct a XLM-RoBERTa tokenizer from a ``curated-tokenizers`` SentencePiece processor.

        :param processor:
            The processor to wrap.
        """
        super().__init__(processor=processor)

        self.processor = processor

        bos_id = processor.bos_id()
        eos_id = processor.eos_id()
        bos_piece = processor.id_to_piece(bos_id)
        eos_piece = processor.id_to_piece(eos_id)

        self.pre_encoder = AddBosEosPreEncoder(bos_piece=bos_piece, eos_piece=eos_piece)
        self.post_encoder = XLMRPostEncoder()
        self.pre_decoder = XLMRPreDecoder(bos_id=bos_id, eos_id=eos_id)

        self._eos_piece = eos_piece

    @classmethod
    def from_files(
        cls: Type[Self],
        *,
        model_file: RepositoryFile,
    ) -> Self:
        """
        Construct a XLM-R tokenizer from a SentencePiece model.

        :param model_file:
            The SentencePiece model file.
        """
        with model_file.open() as f:
            processor = SentencePieceProcessor.from_file(f)
        return cls(processor=processor)

    @classmethod
    def _load_from_vocab_files(
        cls: Type[Self],
        *,
        vocab_files: Mapping[str, RepositoryFile],
        tokenizer_config: Optional[Dict[str, Any]],
    ) -> Self:
        return cls.from_files(model_file=vocab_files["model"])



import pytest
import torch

from curated_transformers.repository.file import LocalFile
from curated_transformers.tokenizers import PiecesWithIds
from curated_transformers.tokenizers.legacy.camembert_tokenizer import (
    CamemBERTTokenizer,
)

from ...compat import has_hf_transformers
from ...utils import torch_assertclose
from ..util import compare_tokenizer_outputs_with_hf_tokenizer


@pytest.fixture
def toy_tokenizer(test_dir):
    return CamemBERTTokenizer.from_files(
        model_file=LocalFile(path=test_dir / "toy.model"),
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_from_hf_hub_equals_hf_tokenizer(sample_texts, french_sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(
        sample_texts, "camembert-base", CamemBERTTokenizer
    )
    compare_tokenizer_outputs_with_hf_tokenizer(
        french_sample_texts, "camembert-base", CamemBERTTokenizer
    )


def test_camembert_tokenizer_toy_tokenizer(toy_tokenizer, short_sample_texts):
    encoding = toy_tokenizer(short_sample_texts)
    _check_toy_tokenizer(encoding)

    decoded = toy_tokenizer.decode(encoding.ids)
    assert decoded == [
        "I saw a girl with a telescope.",
        "Today we will eat pok ⁇  bowl, lots of it!",
        "Tokens which are unknown in ⁇  most ⁇  latin ⁇  alphabet ⁇  vocabularies.",
    ]


def _check_toy_tokenizer(pieces):
    assert isinstance(pieces, PiecesWithIds)
    assert len(pieces.ids) == 3
    assert len(pieces.pieces) == 3

    assert pieces.ids == [
        [5, 12, 469, 14, 951, 45, 14, 174, 172, 114, 32, 24, 147, 8, 6],
        [
            5,
            487,
            550,
            116,
            175,
            571,
            66,
            24,
            49,
            3,
            88,
            119,
            31,
            7,
            152,
            230,
            10,
            17,
            29,
            150,
            6,
        ],
        [
            5,
            487,
            49,
            98,
            10,
            143,
            126,
            225,
            49,
            28,
            119,
            28,
            23,
            3,
            640,
            3,
            152,
            80,
            57,
            3,
            14,
            31,
            33,
            56,
            22,
            69,
            19,
            18,
            3,
            11,
            87,
            24,
            32,
            22,
            69,
            235,
            53,
            461,
            8,
            6,
        ],
    ]

    assert pieces.pieces == [
        [
            "<s>",
            "▁I",
            "▁saw",
            "▁a",
            "▁girl",
            "▁with",
            "▁a",
            "▁t",
            "el",
            "es",
            "c",
            "o",
            "pe",
            ".",
            "</s>",
        ],
        [
            "<s>",
            "▁To",
            "day",
            "▁we",
            "▁will",
            "▁eat",
            "▁p",
            "o",
            "k",
            "é",
            "▁b",
            "ow",
            "l",
            ",",
            "▁l",
            "ot",
            "s",
            "▁of",
            "▁it",
            "!",
            "</s>",
        ],
        [
            "<s>",
            "▁To",
            "k",
            "en",
            "s",
            "▁which",
            "▁are",
            "▁un",
            "k",
            "n",
            "ow",
            "n",
            "▁in",
            "ペ",
            "▁most",
            "で",
            "▁l",
            "at",
            "in",
            "が",
            "▁a",
            "l",
            "p",
            "h",
            "a",
            "b",
            "e",
            "t",
            "際",
            "▁",
            "v",
            "o",
            "c",
            "a",
            "b",
            "ul",
            "ar",
            "ies",
            ".",
            "</s>",
        ],
    ]

    torch_assertclose(
        pieces.padded_tensor(padding_id=1),
        torch.tensor(
            [
                [
                    5,
                    12,
                    469,
                    14,
                    951,
                    45,
                    14,
                    174,
                    172,
                    114,
                    32,
                    24,
                    147,
                    8,
                    6,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
                [
                    5,
                    487,
                    550,
                    116,
                    175,
                    571,
                    66,
                    24,
                    49,
                    3,
                    88,
                    119,
                    31,
                    7,
                    152,
                    230,
                    10,
                    17,
                    29,
                    150,
                    6,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
                [
                    5,
                    487,
                    49,
                    98,
                    10,
                    143,
                    126,
                    225,
                    49,
                    28,
                    119,
                    28,
                    23,
                    3,
                    640,
                    3,
                    152,
                    80,
                    57,
                    3,
                    14,
                    31,
                    33,
                    56,
                    22,
                    69,
                    19,
                    18,
                    3,
                    11,
                    87,
                    24,
                    32,
                    22,
                    69,
                    235,
                    53,
                    461,
                    8,
                    6,
                ],
            ],
            dtype=torch.int32,
        ),
    )
    torch_assertclose(
        pieces.attention_mask().bool_mask.squeeze(dim=(1, 2)),
        torch.tensor(
            [
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
            ]
        ),
    )



import pytest
from huggingface_hub import _CACHED_NO_EXIST, try_to_load_from_cache

from curated_transformers.tokenizers import Tokenizer
from curated_transformers.tokenizers.legacy import BERTTokenizer

from ..compat import has_hf_transformers
from .util import compare_tokenizer_outputs_with_hf_tokenizer


def test_from_hf_hub_to_cache():
    Tokenizer.from_hf_hub_to_cache(
        name="EleutherAI/gpt-neox-20b",
        revision="9369f145ca7b66ef62760f9351af951b2d53b77f",
    )

    expected_files = [
        "tokenizer_config.json",
        "special_tokens_map.json",
        "tokenizer.json",
    ]
    for name in expected_files:
        assert (
            try_to_load_from_cache(
                repo_id="EleutherAI/gpt-neox-20b",
                filename=name,
                revision="9369f145ca7b66ef62760f9351af951b2d53b77f",
            )
            != _CACHED_NO_EXIST
        )


def test_from_hf_hub_to_cache_legacy():
    BERTTokenizer.from_hf_hub_to_cache(
        name="bert-base-uncased",
        revision="1dbc166cf8765166998eff31ade2eb64c8a40076",
    )

    expected_files = [
        "tokenizer_config.json",
        "vocab.txt",
    ]
    for name in expected_files:
        assert (
            try_to_load_from_cache(
                repo_id="bert-base-uncased",
                filename=name,
                revision="1dbc166cf8765166998eff31ade2eb64c8a40076",
            )
            != _CACHED_NO_EXIST
        )


@pytest.mark.xfail(reason="HfFileSystem calls safetensors with incorrect arguments")
@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_fsspec(sample_texts):
    # We only test one model, since using fsspec downloads the model
    # each time.
    compare_tokenizer_outputs_with_hf_tokenizer(
        sample_texts,
        "EleutherAI/gpt-neox-20b",
        Tokenizer,
        pad_token="</s>",
        with_fsspec=True,
    )



from .auto_tokenizer import AutoTokenizer
from .chunks import InputChunks, SpecialPieceChunk, TextChunk
from .hf_hub import FromHF
from .tokenizer import PiecesWithIds, Tokenizer, TokenizerBase

__all__ = [
    "AutoTokenizer",
    "FromHF",
    "InputChunks",
    "PiecesWithIds",
    "SpecialPieceChunk",
    "TextChunk",
    "Tokenizer",
    "TokenizerBase",
]



from abc import ABC, abstractmethod
from typing import Optional, Type, TypeVar

import torch

from ..quantization.bnb.config import BitsAndBytesConfig

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="FromHF")


class FromHF(ABC):
    """
    Mixin class for downloading generators from Hugging Face Hub.

    It automatically infers the tokenizer and model type and loads
    the parameters and the configuration from the hub.
    """

    @classmethod
    @abstractmethod
    def from_hf_hub_to_cache(
        cls: Type[Self],
        *,
        name: str,
        revision: str = "main",
    ):
        """
        Download the generator's model and tokenizer from Hugging Face Hub
        into the local Hugging Face cache directory. Subsequent loading of the
        generator will load the model and the tokenizer from disk.

        :param name:
            Model name.
        :param revision:
            Model revision.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_hf_hub(
        cls: Type[Self],
        *,
        name: str,
        revision: str = "main",
        device: Optional[torch.device] = None,
        quantization_config: Optional[BitsAndBytesConfig] = None,
    ) -> Self:
        """
        Construct a generator and load its parameters from Hugging Face Hub.

        :param name:
            Model name.
        :param revision:
            Model revision.
        :param device:
            Device on which to initialize the model.
        :param quantization_config:
            Configuration for loading quantized weights.
        :returns:
            Generator with the parameters loaded.
        """
        raise NotImplementedError



from typing import Optional

from huggingface_hub import HfFileSystem

from ..compat import transformers
from ..utils import torch_assertclose


def compare_tokenizer_outputs_with_hf_tokenizer(
    sample_texts,
    hf_name,
    tokenizer_cls,
    pad_token: Optional[str] = None,
    with_hf_fast: bool = True,
    with_fsspec: bool = False,
    revision: str = "main",
):
    if with_fsspec:
        tokenizer = tokenizer_cls.from_fsspec(fs=HfFileSystem(), model_path=hf_name)
    else:
        tokenizer = tokenizer_cls.from_hf_hub(name=hf_name, revision=revision)
    pieces = tokenizer(sample_texts)

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained(
        hf_name, revision=revision, use_fast=with_hf_fast
    )
    hf_tokenizer.padding_side = "right"
    if pad_token is not None:
        hf_tokenizer.add_special_tokens({"pad_token": pad_token})

    # Test encoding with right-padding.
    hf_pieces = hf_tokenizer(sample_texts, padding=True, return_tensors="pt")
    torch_assertclose(
        pieces.padded_tensor(padding_id=hf_tokenizer.pad_token_id),
        hf_pieces["input_ids"].int(),
    )
    torch_assertclose(
        pieces.attention_mask().bool_mask.squeeze(dim=(1, 2)),
        hf_pieces["attention_mask"].bool(),
    )

    # Test decoding
    decoded = tokenizer.decode(pieces.ids)
    hf_decoded = [
        hf_tokenizer.decode(ids, skip_special_tokens=True)
        for ids in hf_pieces["input_ids"]
    ]

    assert decoded == hf_decoded

    # Test encoding with left-padding.
    hf_tokenizer.padding_side = "left"
    hf_pieces = hf_tokenizer(sample_texts, padding=True, return_tensors="pt")
    torch_assertclose(
        pieces.padded_tensor(padding_id=hf_tokenizer.pad_token_id, pad_left=True),
        hf_pieces["input_ids"].int(),
    )
    torch_assertclose(
        pieces.attention_mask(pad_left=True).bool_mask.squeeze(dim=(1, 2)),
        hf_pieces["attention_mask"].bool(),
    )



import pytest
import torch

from curated_transformers.generation.logits import (
    TemperatureTransform,
    TopKTransform,
    TopPTransform,
    VocabMaskTransform,
)

from ..utils import torch_assertclose


def test_temperature_transform():
    logits = torch.rand((2, 10, 20), dtype=torch.float)
    transform = TemperatureTransform(temperature=1.0)
    torch_assertclose(transform(logits), logits)
    transform = TemperatureTransform(temperature=0.5)
    torch_assertclose(transform(logits), logits / 0.5)
    transform = TemperatureTransform(temperature=2.0)
    torch_assertclose(transform(logits), logits / 2.0)


def test_invalid_temperature_raises():
    with pytest.raises(ValueError, match=r"The temperature must be a non-zero"):
        TemperatureTransform(temperature=0.0)


def test_top_k_transform():
    logits = torch.tensor(
        [[4.0, 4.0, 3.5, 0.0, 3.5, 0.0, 3.5, 4.0]], dtype=torch.float32
    )
    for k in [1, 2, 3]:
        transform = TopKTransform(k=k)
        torch_assertclose(
            transform(logits),
            torch.tensor(
                [
                    [
                        4.0,
                        4.0,
                        -3.4028e38,
                        -3.4028e38,
                        -3.4028e38,
                        -3.4028e38,
                        -3.4028e38,
                        4.0,
                    ]
                ],
                dtype=torch.float32,
            ),
        )

    for k in [4, 5, 6]:
        transform = TopKTransform(k=k)
        torch_assertclose(
            transform(logits),
            torch.tensor(
                [
                    [
                        4.0,
                        4.0,
                        3.5,
                        -3.4028e38,
                        3.5,
                        -3.4028e38,
                        3.5,
                        4.0,
                    ]
                ],
                dtype=torch.float32,
            ),
        )
        transform = TopKTransform(k=k)

    for k in [0, 8]:
        transform = TopKTransform(k=k)
        torch_assertclose(
            transform(logits),
            torch.tensor(
                [
                    [
                        4.0,
                        4.0,
                        3.5,
                        0.0,
                        3.5,
                        0.0,
                        3.5,
                        4.0,
                    ]
                ],
                dtype=torch.float32,
            ),
        )


def test_top_p_transform():
    logits = torch.tensor(
        [[4.0, 4.0, 3.3, 0.0, 3.5, 0.0, 5.0, 3.4]], dtype=torch.float32
    )

    transform = TopPTransform(0.9)
    torch_assertclose(
        transform(logits),
        torch.tensor([[4.0, 4.0, -3.4028e38, -3.4028e38, 3.5, -3.4028e38, 5.0, 3.4]]),
    )

    # Check p less than the most probable class.
    transform = TopPTransform(0.01)
    torch_assertclose(
        transform(logits),
        torch.tensor(
            [
                [
                    -3.4028e38,
                    -3.4028e38,
                    -3.4028e38,
                    -3.4028e38,
                    -3.4028e38,
                    -3.4028e38,
                    5.0000e00,
                    -3.4028e38,
                ]
            ]
        ),
    )

    # All classes.
    transform = TopPTransform(1.0)
    torch_assertclose(transform(logits), logits)


def test_invalid_top_p_raises():
    with pytest.raises(ValueError, match=r"Top-p.*between"):
        TopPTransform(-0.1)
    with pytest.raises(ValueError, match=r"Top-p.*between"):
        TopPTransform(1.1)


def test_mask_transform():
    logits = torch.tensor(
        [[[4.0, 4.0, 3.5, 0.0, 3.5, 0.0, 3.5, 4.0]]], dtype=torch.float32
    )

    transform = VocabMaskTransform(pieces_to_mask=[2, 4, 5])
    torch_assertclose(
        transform(logits),
        torch.tensor(
            [
                [
                    [
                        4.0,
                        4.0,
                        -3.4028e38,
                        0.0,
                        -3.4028e38,
                        -3.4028e38,
                        3.5,
                        4.0,
                    ]
                ]
            ],
            dtype=torch.float32,
        ),
    )

    transform = VocabMaskTransform(pieces_to_mask=[])
    torch_assertclose(transform(logits, inplace=False), logits)


def test_invalid_mask_transform_raises():
    with pytest.raises(ValueError, match="must be 1D"):
        transform = VocabMaskTransform(pieces_to_mask=[[1, 2]])

    with pytest.raises(ValueError, match="must be >= 0"):
        transform = VocabMaskTransform(pieces_to_mask=[-1, 0])

    with pytest.raises(ValueError, match="must be < .*, but got"):
        transform = VocabMaskTransform(pieces_to_mask=[8])
        transform(
            torch.tensor(
                [[4.0, 4.0, 3.5, 0.0, 3.5, 0.0, 3.5, 4.0]], dtype=torch.float32
            )
        )



from typing import Any, Dict, List, Mapping, Optional, Set, Tuple, Type, TypeVar

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn import Embedding

from ...layers.attention import AttentionMask
from ...layers.cache import KeyValueCache
from ...quantization import Quantizable
from ..hf_hub import FromHF
from ..hf_hub.conversion import state_dict_from_hf, state_dict_to_hf
from ..output import CausalLMOutputWithCache
from ..transformer import TransformerCausalLM
from ._hf import CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS, _config_from_hf, _config_to_hf
from .config import MPTConfig
from .decoder import MPTDecoder

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="MPTCausalLM")


class MPTCausalLM(TransformerCausalLM[MPTConfig], FromHF[MPTConfig], Quantizable):
    """
    `MosaicML MPT`_ causal language model.

    .. _MosaicML MPT: https://www.mosaicml.com/blog/mpt-7b
    """

    def __init__(
        self, config: MPTConfig, *, device: Optional[torch.device] = None
    ) -> None:
        """
        Construct an MPT causal LM.

        :param config:
            Causal LM configuration.
        :param device:
            Device to which the module is to be moved.
        :returns:
            The causal LM.
        """
        super().__init__(config)

        self.decoder = MPTDecoder(config, device=device)

        # Once we have proper support for tied weights, we will do something like:
        #
        # self.output_embeddings = Linear(
        #    in_features=config.layer.feedforward.hidden_width,
        #    out_features=config.embedding.n_pieces,
        #    bias=False,
        #    device=device,
        # )
        # self.output_embeddings.weights = self.decoder.embeddings.piece_embeddings.weights
        #
        # For now we'll work around this by using the piece embeddings directly.

    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        cache: Optional[List[KeyValueCache]] = None,
        positions: Optional[Tensor] = None,
        store_cache: bool = False,
    ) -> CausalLMOutputWithCache[KeyValueCache]:
        # TODO: remove this forward method once we support weight tying.

        decoder_output = self.decoder(
            piece_ids,
            attention_mask,
            cache=cache,
            store_cache=store_cache,
            positions=positions,
        )

        assert isinstance(self.decoder.embeddings.piece_embeddings, Embedding)
        output_embeddings = self.decoder.embeddings.piece_embeddings.weight

        logits = F.linear(decoder_output.last_hidden_layer_state, output_embeddings)
        return CausalLMOutputWithCache(
            all_outputs=decoder_output.all_outputs,
            cache=decoder_output.cache,
            logits=logits,
        )

    @classmethod
    def is_supported(cls: Type[Self], config: Dict[str, Any]) -> bool:
        return config.get("model_type") == "mpt"

    @classmethod
    def state_dict_from_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_from_hf(params, CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def state_dict_to_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_to_hf(params, CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> MPTConfig:
        return _config_from_hf(hf_config)

    @classmethod
    def config_to_hf(cls, curated_config: MPTConfig) -> Mapping[str, Any]:
        return _config_to_hf(cls, curated_config)

    @classmethod
    def from_hf_config(
        cls: Type[Self],
        *,
        hf_config: Any,
        device: Optional[torch.device] = None,
    ) -> Self:
        config = cls.config_from_hf(hf_config)
        return cls(config, device=device)

    @classmethod
    def modules_to_not_quantize(cls) -> Set[str]:
        # Ignore the output embedding matrix.
        return {"output_embeddings"}



from dataclasses import dataclass

import torch

from ...layers.activations import Activation
from ..config import (
    RotaryEmbeddingConfig,
    TransformerAttentionLayerConfig,
    TransformerConfig,
    TransformerEmbeddingLayerConfig,
    TransformerFeedForwardLayerConfig,
    TransformerLayerConfig,
)


@dataclass
class GPTNeoXConfig(TransformerConfig):
    """
    GPT-NeoX (`Black et al., 2022`_) model configuration.

    .. _Black et al., 2022: https://arxiv.org/abs/2204.06745
    """

    def __init__(
        self,
        *,
        attention_probs_dropout_prob: float = 0.0,
        activation: Activation = Activation.GELU,
        dtype: torch.dtype = torch.float16,
        hidden_dropout_prob: float = 0.0,
        hidden_width: int = 2560,
        intermediate_width: int = 10240,
        layer_norm_eps: float = 1e-5,
        n_positions: int = 2048,
        model_max_length: int = 2048,
        n_attention_heads: int = 32,
        n_hidden_layers: int = 32,
        rotary_embedding_base: int = 10000,
        rotary_embedding_fraction: float = 0.25,
        n_pieces: int = 50280,
    ):
        """
        :param attention_probs_dropout_prob:
            Dropout to apply after attention.
        :param activation:
            Activation used by the pointwise feed-forward layers.
        :param dtype:
            Data type to use for model parameters.
        :param hidden_dropout_prob:
            Dropout to apply to the hidden and embedding layers.
        :param hidden_width:
            Hidden width of the transformer.
        :param intermediate_width:
            Intermediate width in the feed-forward layer. The non-linearity
            is applied in this intermediate width.
        :param layer_norm_eps:
            Epsilon for layer normalization.
        :param n_attention_heads:
            Number of attention heads.
        :param n_hidden_layers:
            Number of hidden layers.
        :param rotary_embedding_base:
            Base in signifying the rotary embedding period.
        :param rotary_embedding_fraction:
            Fraction of hidden width to apply rotary embeddings to.
            Must be in ``[0,1]``.
        :param n_pieces:
            Vocabulary size (number of embeddings).
        """

        # TODO: n_positions and model_max_length are currently
        #       not used. We may want to limit the rotary embeddings to these
        #       values in the future. We should check empirically if the auto
        #       resizing in rotary embeddings makes sense.

        self.embedding = TransformerEmbeddingLayerConfig(
            dropout_prob=hidden_dropout_prob,
            embedding_width=hidden_width,
            n_pieces=n_pieces,
            layer_norm_eps=layer_norm_eps,
            n_positions=None,
            n_types=None,
        )
        self.layer = TransformerLayerConfig(
            attention=TransformerAttentionLayerConfig(
                dropout_prob=attention_probs_dropout_prob,
                hidden_width=hidden_width,
                n_query_heads=n_attention_heads,
                n_key_value_heads=n_attention_heads,
                rotary_embeddings=RotaryEmbeddingConfig(
                    rotary_fraction=rotary_embedding_fraction,
                    rotary_base=rotary_embedding_base,
                ),
                use_alibi=False,
                use_bias=True,
                use_parallel_attention=True,
            ),
            feedforward=TransformerFeedForwardLayerConfig(
                hidden_width=hidden_width,
                intermediate_width=intermediate_width,
                activation=activation,
                use_bias=True,
                use_gate=False,
            ),
            dropout_prob=hidden_dropout_prob,
            layer_norm_eps=layer_norm_eps,
            n_hidden_layers=n_hidden_layers,
        )
        self.dtype = dtype



import pytest
import torch

from curated_transformers.generation.config import (
    GreedyGeneratorConfig,
    SampleGeneratorConfig,
)
from curated_transformers.generation.default_generator import DefaultGenerator
from curated_transformers.generation.logits import VocabMaskTransform

from ..conftest import GPU_TESTS_ENABLED


@pytest.fixture(scope="module")
def falcon_generator():
    return DefaultGenerator.from_hf_hub(
        name="tiiuae/falcon-7b-instruct", device=torch.device("cuda", index=0)
    )


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
def test_generate_deterministic(falcon_generator):
    prompts = [
        "What is the Rust programming language?\n",
        "What is spaCy?\n",
    ]
    answers = [
        "Rust is a programming language that is designed to be a safe, concurrent, and efficient replacement for C++. It is a statically-typed language that is designed to be memory-safe and thread-safe, making it a good choice for developing high-performance applications.",
        "spaCy is a Python library for natural language processing. It is designed to be easy to use and highly customizable, making it a great tool for developers and researchers.",
    ]
    assert falcon_generator(prompts, config=GreedyGeneratorConfig()) == answers

    # Test in reverse order to verify that sequence identifiers are
    # handled correctly.
    assert (
        falcon_generator(prompts[::-1], config=GreedyGeneratorConfig()) == answers[::-1]
    )


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
def test_generate_max_generated_pieces(falcon_generator):
    prompts = [
        "What is the Rust programming language?\n",
        "What is spaCy?\n",
    ]
    answers = [
        "Rust is a programming language that is designed to",
        "spaCy is a Python library for natural language",
    ]
    assert (
        falcon_generator(prompts, config=GreedyGeneratorConfig(max_generated_pieces=10))
        == answers
    )

    # Test in reverse order to verify that sequence identifiers are
    # handled correctly.
    assert (
        falcon_generator(
            prompts[::-1], config=GreedyGeneratorConfig(max_generated_pieces=10)
        )
        == answers[::-1]
    )


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
def test_generate_sample(falcon_generator):
    prompts = [
        "What is spaCy?\n",
        "What is spaCy?\n",
    ]
    # Fix the seed so that we are always randomly sampling in the same way.
    torch.manual_seed(0)
    assert falcon_generator(prompts, config=SampleGeneratorConfig(top_k=10)) == [
        "spaCy is a Python package for natural language processing and text analysis. It is specifically designed for text classification tasks and can be used in a variety of fields, including healthcare, finance, and marketing. spaCy's main feature is its ability to extract text and its associated entities, which are then used to perform various analyses on the text, such as sentiment analysis, named entity recognition, and entity extraction.",
        "spaCy is a library for natural language processing in Python. It's built on top of NLTK's WordNet, and uses it to create a set of spaCy-compatible data structures for representing words and phrases in text. It's also a grammar-based approach to word and phrase similarity matching, and provides an API for building custom grammars.",
    ]

    torch.manual_seed(0)
    assert falcon_generator(
        prompts, config=SampleGeneratorConfig(top_k=5, temperature=2)
    ) == [
        "spacy is a Natural Language Processing tool that can be used with many programming language to build NLP-based applications such as machine learning, sentiment analysis and chatbots, and to extract text from documents and other media. spacy uses the Stanford NLP model to learn from and generate text. This makes it one of the most popular and versatile NLP libraries available. The main features of spacy include text generation and manipulation, entity extraction, part-of-speech tagging and more.",
        "spaCy is a library for natural language processing in Scala. It's designed to be easy to use and provides a range of features including text analysis tools and pre-built models to analyze various text datasets.</s> \nCan spaCy be used to analyze text from different sources or does it only work with a specific type of text or data?</s> \nspaCy can analyze text from different sources and can be used to work with text in different languages such as German and Chinese.",
    ]


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
def test_generate_masked_output(falcon_generator):
    prompt = [
        "Repeat the following word: Madagascar\n",
    ]
    pieces_to_mask = [
        58588,  # ĠMadagascar
        18704,  # Mad
        54453,  # agascar
    ]

    generated = falcon_generator(
        prompt,
        config=GreedyGeneratorConfig(
            max_generated_pieces=50,
            masked_pieces=pieces_to_mask,
        ),
    )
    assert "Madagascar" not in generated



from typing import Generic, List, Optional, Tuple

import torch
from torch import Tensor

from ..layers.attention import AttentionMask
from ..models.output import CacheT
from .stop_conditions import StopCondition


class GeneratorState(Generic[CacheT]):
    """
    Stores the state of the generation process and tracks
    the sequences being generated.
    """

    attention_mask: AttentionMask
    cache: Optional[List[CacheT]]
    positions: Tensor
    seq_ids: Tensor
    prompt_ids: Tensor
    generated_ids: Tensor

    def __init__(
        self,
        *,
        attention_mask: AttentionMask,
        cache: Optional[List[CacheT]],
        prompt_ids: Tensor,
    ) -> None:
        """
        Construct a generator state.

        :param attention_mask:
            Attention mask for the prompts.
        :param cache:
            Transformer model cache.
        :param prompt_ids:
            Batch of prompts.

            *Shape:* ``(batch_size, seq_len)``
        """
        device = prompt_ids.device
        assert (
            attention_mask.device == device
        ), f"Attention mask device '{attention_mask.device}' is not same as prompt ids device '{prompt_ids.device}'"
        self.attention_mask = attention_mask
        self.positions = attention_mask.bool_mask.int().cumsum(-1) - 1
        self.cache = cache
        self.seq_ids = torch.arange(0, self.attention_mask.shape[0], device=device)
        self.prompt_ids = prompt_ids
        self.generated_ids = torch.zeros(
            (prompt_ids.size(0), 0), dtype=prompt_ids.dtype, device=device
        )

    @property
    def is_finished(self):
        """
        Whether all sequences have finished generating.

        :returns:
            ``True`` iff all sequences have finished generating.
        """
        return len(self.seq_ids) == 0

    @property
    def last_step_ids(self) -> Tensor:
        """
        Identifiers generated in the last step.

        :returns:
            Generated identifiers. Prompt identifiers are returned
            when the generator has not stepped yet.
        """
        if not self.generated_ids.size(1):
            return self.prompt_ids
        else:
            return self.generated_ids[:, -1:]

    def step(
        self,
        *,
        cache: List[CacheT],
        predicted_ids: Tensor,
        stop_condition: StopCondition,
    ) -> Tuple[Tensor, Tensor]:
        """
        Step the generation state.

        :param cache:
            Model cache from the last model call.
        :param generated_ids:
            Tensor containing generated IDs.

            *Shape:* ``(batch_size, 1)``
        :param stop_condition:
            Generation stop condition.
        :returns:
            Sequence identifiers and piece IDs.

            *Shape:* ``(batch_size), (batch_size, 1)``
        """
        # We update the state before removing completed sequences, so that
        # stopping conditions get a consistent view.
        self.cache = cache
        self.generated_ids = torch.concat([self.generated_ids, predicted_ids], 1)
        self.attention_mask = self.attention_mask.extend_length(
            count=1, fill_value=True
        )
        self.positions = self.positions.max(-1, keepdim=True).values + 1

        # Determine which sequences are done generating and remove them.
        completed_exclude = torch.zeros_like(predicted_ids, dtype=torch.bool)
        completed_include = torch.zeros_like(predicted_ids, dtype=torch.bool)
        stop_condition.update_completed(
            state=self,
            completed_exclude=completed_exclude,
            completed_include=completed_include,
        )

        # Prepare sequences and identifiers that the generator should yield.
        to_yield = completed_exclude.logical_not().view(-1)
        seq_ids = self.seq_ids[to_yield.view(-1)]
        last_step_ids = self.last_step_ids[to_yield]

        self._remove_completed((completed_exclude ^ completed_include).view(-1))

        return seq_ids, last_step_ids

    def _remove_completed(self, completed: Tensor):
        """
        Remove completed sequences.

        :param completed:
            Tensor indicating for the active sequences whether they are completed.

        :meta private:
        """
        not_completed = completed.logical_not()
        self.generated_ids = self.generated_ids[not_completed]
        self.attention_mask = self.attention_mask.filter_batch_items(not_completed)
        if self.cache is not None:
            self.cache = [
                layer_cache.filter_batch_items(not_completed)
                for layer_cache in self.cache
            ]
        self.prompt_ids = self.prompt_ids[not_completed]
        self.positions = self.positions[not_completed]
        self.seq_ids = self.seq_ids[not_completed]



from typing import Any, Dict, Mapping, Optional, Type, TypeVar

from curated_tokenizers import SentencePieceProcessor

from ...repository.file import RepositoryFile
from ..hf_hub import LegacyFromHF
from .legacy_tokenizer import AddBosEosPreEncoder
from .sentencepiece_tokenizer import SentencePieceTokenizer

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="LlamaTokenizer")

DEFAULT_BOS_PIECE = "<s>"


class LlamaTokenizer(SentencePieceTokenizer, LegacyFromHF):
    """
    Legacy tokenizer for Llama (`Touvron et al., 2023 [a]`_, `Touvron et al., 2023 [b]`_) models.

    .. _Touvron et al., 2023 [a]: https://arxiv.org/abs/2302.13971
    .. _Touvron et al., 2023 [b]: https://arxiv.org/abs/2307.09288
    """

    vocab_files: Dict[str, str] = {"model": "tokenizer.model"}

    def __init__(
        self,
        *,
        processor: SentencePieceProcessor,
        add_bos_piece: bool = True,
        add_eos_piece: bool = False,
    ):
        """
        Construct a Llama tokenizer from a ``curated-tokenizers`` SentencePiece processor.

        :param processor:
            The processor to wrap.
        :param add_bos_piece:
            Add a begin-of-sequence piece.
        :param add_eos_piece:
            Add an end-of-sequence piece.
        """
        super().__init__(processor=processor)

        self.processor = processor

        bos_piece = processor.id_to_piece(processor.bos_id()) if add_bos_piece else None
        eos_piece = processor.id_to_piece(processor.eos_id()) if add_eos_piece else None

        self.pre_encoder = AddBosEosPreEncoder(bos_piece=bos_piece, eos_piece=eos_piece)

    @classmethod
    def from_files(
        cls: Type[Self],
        *,
        model_file: RepositoryFile,
        add_bos_piece: bool = True,
        add_eos_piece: bool = False,
    ) -> Self:
        """
        Construct a Llama tokenizer from a SentencePiece model.

        :param model_file:
            The SentencePiece model file.
        :param add_bos_piece:
            Add a begin-of-sequence piece.
        :param add_eos_piece:
            Add an end-of-sequence piece.
        """
        with model_file.open() as f:
            processor = SentencePieceProcessor.from_file(f)
        return cls(
            processor=processor,
            add_bos_piece=add_bos_piece,
            add_eos_piece=add_eos_piece,
        )

    @classmethod
    def _load_from_vocab_files(
        cls: Type[Self],
        *,
        vocab_files: Mapping[str, RepositoryFile],
        tokenizer_config: Optional[Dict[str, Any]],
    ) -> Self:
        if tokenizer_config is None:
            return cls.from_files(model_file=vocab_files["model"])

        add_bos_piece = tokenizer_config.get("add_bos_token", True)
        add_eos_piece = tokenizer_config.get("add_eos_token", False)

        return cls.from_files(
            model_file=vocab_files["model"],
            add_bos_piece=add_bos_piece,
            add_eos_piece=add_eos_piece,
        )



from typing import Any, Dict, List, Mapping, Optional, Tuple

import torch

from ...util.string import StringTransform, StringTransformations
from ..hf_hub.conversion import (
    CommonCuratedToHFConverters,
    CommonHFKeys,
    HFConfigKey,
    HFConfigKeyDefault,
    HFSpecificConfig,
    config_from_hf,
    config_to_hf,
)
from ..module import DecoderModule
from .config import FalconConfig

ATTENTION_DROPOUT = "attention_probs_dropout_prob"
HIDDEN_DROPOUT = "hidden_dropout_prob"
EXTRA_KWARG_KEYS = [ATTENTION_DROPOUT, HIDDEN_DROPOUT]


# Order-dependent.
COMMON_HF_PARAM_KEY_TRANSFORMS: List[StringTransform] = [
    StringTransformations.regex_sub((r"^h\.", "layers."), (r"^layers\.", "h.")),
    StringTransformations.sub("decoder.h.", "decoder.layers."),
    # Attention blocks.
    StringTransformations.sub(".self_attention", ".mha"),
    StringTransformations.sub(".mha.query_key_value", ".mha.input"),
    StringTransformations.sub(".mha.dense", ".mha.output"),
    # Pointwise feedforward.
    StringTransformations.sub(".mlp", ".ffn"),
    StringTransformations.sub(".dense_h_to_4h", ".intermediate"),
    StringTransformations.sub(".ffn.dense_4h_to_h", ".ffn.output"),
    # Layer norms.
    StringTransformations.sub(".input_layernorm", ".attn_layer_norm"),
    StringTransformations.sub(".ln_attn", ".attn_input_layer_norm"),
    StringTransformations.sub(".post_attention_layernorm", ".ffn_layer_norm"),
    StringTransformations.sub(".ln_mlp", ".ffn_input_layer_norm"),
    StringTransformations.sub("ln_f.", "output_layer_norm."),
    # Embeddings.
    StringTransformations.sub("word_embeddings.", "embeddings.piece_embeddings."),
    StringTransformations.sub("lm_head.", "output_embeddings."),
]

DECODER_HF_PARAM_KEY_TRANSFORMS = [
    StringTransformations.remove_prefix("transformer.", reversible=False)
] + COMMON_HF_PARAM_KEY_TRANSFORMS
CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS = [
    StringTransformations.sub("transformer.", "decoder."),
] + COMMON_HF_PARAM_KEY_TRANSFORMS


class HFConfigKeys:
    @staticmethod
    def conv_multi_query(config: FalconConfig) -> bool:
        return (
            config.layer.attention.n_query_heads
            != config.layer.attention.n_key_value_heads
        )

    @staticmethod
    def conv_n_attention_query_heads(config: FalconConfig) -> int:
        return config.layer.attention.n_query_heads

    @staticmethod
    def conv_n_attention_keyvalue_heads(config: FalconConfig) -> int:
        return config.layer.attention.n_key_value_heads

    @staticmethod
    def conv_use_bias(config: FalconConfig) -> bool:
        return config.layer.feedforward.use_bias

    @staticmethod
    def conv_use_alibi(config: FalconConfig) -> bool:
        return config.layer.attention.use_alibi

    @staticmethod
    def conv_use_parallel_attention(config: FalconConfig) -> bool:
        return config.layer.attention.use_parallel_attention

    @staticmethod
    def conv_new_decoder_architecture(config: FalconConfig) -> bool:
        return config.new_decoder_architecture

    # Used by Refined Web Model.
    N_LAYER = HFConfigKey(
        "n_layer",
        "n_hidden_layers",
        lambda c: CommonCuratedToHFConverters.n_hidden_layers(c),
    )
    N_HEAD = HFConfigKey(
        "n_head",
        "n_query_heads",
        lambda c: HFConfigKeys.conv_n_attention_query_heads(c),
    )
    N_HEAD_KV = HFConfigKey(
        "n_head_kv",
        "n_key_value_heads",
        lambda c: HFConfigKeys.conv_n_attention_keyvalue_heads(c),
    )
    # Used by Falcon.
    NUM_HEAD_KV = HFConfigKey(
        "num_kv_heads",
        "n_key_value_heads",
        lambda c: HFConfigKeys.conv_n_attention_keyvalue_heads(c),
    )
    NUM_ATTENTION_HEADS = HFConfigKey(
        "num_attention_heads",
        "n_query_heads",
        lambda c: HFConfigKeys.conv_n_attention_query_heads(c),
    )
    # Used by both.
    LAYER_NORM_EPSILON = HFConfigKey(
        "layer_norm_epsilon",
        "layer_norm_eps",
        lambda c: CommonCuratedToHFConverters.layer_norm_eps(c),
    )
    BIAS = HFConfigKey(
        "bias",
        "use_bias",
        lambda c: HFConfigKeys.conv_use_bias(c),
    )
    ALIBI = HFConfigKey(
        "alibi",
        "use_alibi",
        lambda c: HFConfigKeys.conv_use_alibi(c),
    )
    PARALLEL_ATTN = HFConfigKey(
        "parallel_attn",
        "use_parallel_attention",
        lambda c: HFConfigKeys.conv_use_parallel_attention(c),
    )
    NEW_DECODER_ARCHITECTURE = HFConfigKey(
        "new_decoder_architecture",
        "new_decoder_architecture",
        lambda c: HFConfigKeys.conv_new_decoder_architecture(c),
    )
    # The following keys are not directly converted to kwargs
    # but still need to serialized back to the HF config.
    MULTI_QUERY = HFConfigKey(
        "multi_query",
        "multi_query",
        lambda c: HFConfigKeys.conv_multi_query(c),
    )


# Corresponds to the implementation that the Falcon models
# originally shipped with.
HF_CONFIG_KEYS_REFINED_WEB_MODEL: List[
    Tuple[HFConfigKey, Optional[HFConfigKeyDefault]]
] = [
    (CommonHFKeys.DTYPE, HFConfigKeyDefault("bfloat16")),
    (CommonHFKeys.HIDDEN_SIZE, None),
    (HFConfigKeys.N_HEAD, None),
    (HFConfigKeys.N_HEAD_KV, HFConfigKeyDefault(-1)),
    (HFConfigKeys.MULTI_QUERY, HFConfigKeyDefault(False)),
    (HFConfigKeys.N_LAYER, None),
    (HFConfigKeys.PARALLEL_ATTN, None),
    (HFConfigKeys.BIAS, None),
    (HFConfigKeys.ALIBI, None),
    (HFConfigKeys.LAYER_NORM_EPSILON, None),
    (CommonHFKeys.VOCAB_SIZE, None),
    (CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
    (CommonHFKeys.HIDDEN_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
]

# Corresponds to the mainline implementation for Falcon models
# in the `transformers` library.
HF_CONFIG_KEYS_FALCON: List[Tuple[HFConfigKey, Optional[HFConfigKeyDefault]]] = [
    (CommonHFKeys.DTYPE, HFConfigKeyDefault("bfloat16")),
    (CommonHFKeys.HIDDEN_SIZE, None),
    (HFConfigKeys.NUM_ATTENTION_HEADS, None),
    (HFConfigKeys.NUM_HEAD_KV, HFConfigKeyDefault(-1)),
    (HFConfigKeys.MULTI_QUERY, HFConfigKeyDefault(False)),
    (CommonHFKeys.NUM_HIDDEN_LAYERS, None),
    (HFConfigKeys.PARALLEL_ATTN, None),
    (HFConfigKeys.BIAS, None),
    (HFConfigKeys.ALIBI, None),
    (HFConfigKeys.LAYER_NORM_EPSILON, None),
    (CommonHFKeys.VOCAB_SIZE, None),
    (HFConfigKeys.NEW_DECODER_ARCHITECTURE, HFConfigKeyDefault(False)),
    (CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
    (CommonHFKeys.HIDDEN_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
]


HF_SPECIFIC_CONFIG_DECODER = HFSpecificConfig(
    architectures=["FalconModel"], model_type="falcon"
)
HF_SPECIFIC_CONFIG_CAUSAL_LM = HFSpecificConfig(
    architectures=["FalconForCausalLM"], model_type="falcon"
)


def _config_from_hf_refined_web_model(hf_config: Mapping[str, Any]) -> FalconConfig:
    kwargs = config_from_hf("Falcon", hf_config, HF_CONFIG_KEYS_REFINED_WEB_MODEL)

    # For the old configuration format, we can only figure out whether to use
    # the new decoder architecture by checking if `n_head_kv` is present.
    new_decoder_architecture = HFConfigKeys.N_HEAD_KV.get_kwarg(kwargs) != -1
    multi_query = HFConfigKeys.MULTI_QUERY.get_kwarg(kwargs)

    HFConfigKeys.NEW_DECODER_ARCHITECTURE.set_kwarg(new_decoder_architecture, kwargs)
    HFConfigKeys.MULTI_QUERY.remove_kwarg(kwargs)

    if not new_decoder_architecture:
        if multi_query:
            HFConfigKeys.N_HEAD_KV.set_kwarg(1, kwargs)
        else:
            HFConfigKeys.N_HEAD_KV.set_kwarg(
                HFConfigKeys.N_HEAD.get_kwarg(kwargs), kwargs
            )

    return FalconConfig(
        rotary_embedding_base=10000,
        rotary_embedding_fraction=1.0,
        **kwargs,
    )


def _config_from_hf_falcon(hf_config: Mapping[str, Any]) -> FalconConfig:
    kwargs = config_from_hf("Falcon", hf_config, HF_CONFIG_KEYS_FALCON)

    new_decoder_architecture = HFConfigKeys.NEW_DECODER_ARCHITECTURE.get_kwarg(kwargs)
    multi_query = HFConfigKeys.MULTI_QUERY.get_kwarg(kwargs)

    HFConfigKeys.MULTI_QUERY.remove_kwarg(kwargs)

    if new_decoder_architecture:
        if HFConfigKeys.NUM_HEAD_KV.get_kwarg(kwargs) == -1:
            raise ValueError(
                f"Hugging Face Falcon config with new decoder architecture must contain `n_head_kv`"
            )
    else:
        if multi_query:
            HFConfigKeys.NUM_HEAD_KV.set_kwarg(1, kwargs)
        else:
            HFConfigKeys.NUM_HEAD_KV.set_kwarg(
                HFConfigKeys.NUM_ATTENTION_HEADS.get_kwarg(kwargs), kwargs
            )

    return FalconConfig(
        rotary_embedding_base=10000,
        rotary_embedding_fraction=1.0,
        **kwargs,
    )


def _config_from_hf(hf_config: Mapping[str, Any]) -> FalconConfig:
    model_type = hf_config["model_type"]
    if "RefinedWeb" in model_type:
        return _config_from_hf_refined_web_model(hf_config)
    elif model_type == "falcon":
        return _config_from_hf_falcon(hf_config)
    else:
        raise ValueError(f"Unknown type of Falcon model: {model_type}")


def _config_to_hf(cls, curated_config: FalconConfig) -> Dict[str, Any]:
    # We only support exporting as the mainline Falcon implementation.
    out = config_to_hf(curated_config, [k for k, _ in HF_CONFIG_KEYS_FALCON])
    if issubclass(cls, DecoderModule):
        return HF_SPECIFIC_CONFIG_DECODER.merge(out)
    else:
        return HF_SPECIFIC_CONFIG_CAUSAL_LM.merge(out)



from .auto_generator import AutoGenerator
from .config import GeneratorConfig, GreedyGeneratorConfig, SampleGeneratorConfig
from .default_generator import DefaultGenerator
from .dolly_v2 import DollyV2Generator
from .falcon import FalconGenerator
from .generator import Generator
from .generator_wrapper import GeneratorWrapper
from .hf_hub import FromHF
from .llama import LlamaGenerator
from .logits import (
    CompoundLogitsTransform,
    LogitsTransform,
    TemperatureTransform,
    TopKTransform,
    TopPTransform,
    VocabMaskTransform,
)
from .mpt import MPTGenerator
from .stop_conditions import (
    CompoundStopCondition,
    EndOfSequenceCondition,
    MaxGeneratedPiecesCondition,
    StopCondition,
)
from .string_generator import StringGenerator

__all__ = [
    "AutoGenerator",
    "CompoundLogitsTransform",
    "CompoundStopCondition",
    "DefaultGenerator",
    "DollyV2Generator",
    "EndOfSequenceCondition",
    "FalconGenerator",
    "FromHF",
    "Generator",
    "GeneratorConfig",
    "GeneratorWrapper",
    "GreedyGeneratorConfig",
    "LlamaGenerator",
    "LogitsTransform",
    "MaxGeneratedPiecesCondition",
    "MPTGenerator",
    "SampleGeneratorConfig",
    "StopCondition",
    "StringGenerator",
    "TemperatureTransform",
    "TopKTransform",
    "TopPTransform",
    "VocabMaskTransform",
]



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional, Set

from .logits import (
    CompoundLogitsTransform,
    LogitsTransform,
    TemperatureTransform,
    TopKTransform,
    TopPTransform,
    VocabMaskTransform,
)
from .stop_conditions import (
    CompoundStopCondition,
    EndOfSequenceCondition,
    MaxGeneratedPiecesCondition,
    StopCondition,
)


@dataclass
class GeneratorConfig(ABC):
    """
    Configuration of the generator.

    :param masked_pieces:
        Vocabulary pieces that should be masked out.
    :param eos_id:
        End-of-sequence identifier that should end the generation of a sequence
        when predicted. When this value is set to `None`, it is the
        responsibility of the generator to set it.
    :param max_generated_pieces:
        The maximum number of generation steps. This condition is a noop
        for values less than 1. When this value is set to `None`, it is the
        responsibility of the generator to set it.
    """

    masked_pieces: Optional[Set[int]] = None
    eos_id: Optional[int] = None
    max_generated_pieces: Optional[int] = None

    @abstractmethod
    def logits_transform(self) -> LogitsTransform:
        """
        Get logit transform for the configuration.

        :returns:
            Logits transform. Usually multiple composed transforms.
        """
        raise NotImplementedError

    def stop_condition(self) -> StopCondition:
        """
        Get the stop condition for the configuration.

        :returns:
            Stop condition. Usually multiple composed conditions.
        """
        conditions = CompoundStopCondition()
        if self.eos_id is None:
            raise ValueError("End-of-sequence piece id is unset")
        else:
            conditions.append(EndOfSequenceCondition(self.eos_id))

        if self.max_generated_pieces is None:
            raise ValueError("Maximum number of generation steps is unset")
        else:
            conditions.append(
                MaxGeneratedPiecesCondition(
                    max_generated_pieces=self.max_generated_pieces
                )
            )

        return conditions


@dataclass
class GreedyGeneratorConfig(GeneratorConfig):
    """
    Configuration for greedy generation.

    Greedy generation always selects the highest-probability piece, leading
    to deterministic generation.
    """

    def logits_transform(self) -> LogitsTransform:
        if self.masked_pieces is not None:
            return VocabMaskTransform(self.masked_pieces)
        else:
            return CompoundLogitsTransform([])


@dataclass
class SampleGeneratorConfig(GeneratorConfig):
    """
    Configuration for generation with sampling.

    Sampling-based generation samples pieces from probability distributions.
    Generation is non-deterministic as a result, but provides more varied
    output.

    :param temperature:
        Softmax temperature. For a temperature ``T``:

        - ``T = 1``: the distribution is not changed.
        - ``T < 1``: the entropy of the distribution is decreased.
        - ``T > 1``: the entropy of the distribution is increased.
    :param top_k:
        Sample from top-k highest-probability pieces. ``top_k < 1`` disables top-k
        filtering.
    :param top_p:
        Sample from highest probability pieces the smallest set, such that their
        cumulative probability is >= p. ``top_p = 1.0`` disables top-p filtering.
    """

    temperature: float = 1.0
    top_k: int = 0
    top_p: float = 1.0

    def logits_transform(self) -> LogitsTransform:
        transforms: List[LogitsTransform] = []
        if self.masked_pieces is not None:
            transforms.append(VocabMaskTransform(self.masked_pieces))
        transforms += [
            TemperatureTransform(self.temperature),
            TopKTransform(self.top_k),
            TopPTransform(self.top_p),
        ]

        return CompoundLogitsTransform(transforms)



from .bbpe_tokenizer import ByteBPETokenizer
from .bert_tokenizer import BERTTokenizer
from .camembert_tokenizer import CamemBERTTokenizer
from .legacy_tokenizer import (
    LegacyTokenizer,
    PostDecoder,
    PostEncoder,
    PreDecoder,
    PreEncoder,
)
from .llama_tokenizer import LlamaTokenizer
from .roberta_tokenizer import RoBERTaTokenizer
from .sentencepiece_tokenizer import SentencePieceTokenizer
from .wordpiece_tokenizer import WordPieceTokenizer
from .xlmr_tokenizer import XLMRTokenizer

__all__ = [
    "BERTTokenizer",
    "ByteBPETokenizer",
    "CamemBERTTokenizer",
    "LlamaTokenizer",
    "LegacyTokenizer",
    "PostDecoder",
    "PostEncoder",
    "PreDecoder",
    "PreEncoder",
    "RoBERTaTokenizer",
    "SentencePieceTokenizer",
    "WordPieceTokenizer",
    "XLMRTokenizer",
]



from typing import Any, Dict, Mapping

from ..hf_hub.conversion import HFSpecificConfig
from ..roberta._hf import _config_from_hf as _roberta_config_from_hf
from ..roberta._hf import _config_to_hf as _roberta_config_to_hf
from ..roberta.config import RoBERTaConfig

HF_SPECIFIC_CONFIG = HFSpecificConfig(
    architectures=["CamembertModel"], model_type="camembert"
)


def _config_from_hf(hf_config: Mapping[str, Any]) -> RoBERTaConfig:
    return _roberta_config_from_hf(hf_config)


def _config_to_hf(curated_config: RoBERTaConfig) -> Dict[str, Any]:
    out = _roberta_config_to_hf(curated_config)
    return HF_SPECIFIC_CONFIG.merge(out)



from typing import Any, Dict, List, Mapping, Optional, Tuple

import torch

from ...util.string import StringTransform, StringTransformations
from ..hf_hub.conversion import (
    CommonCuratedToHFConverters,
    CommonHFKeys,
    HFConfigKey,
    HFConfigKeyDefault,
    HFSpecificConfig,
    config_from_hf,
    config_to_hf,
)
from ..module import DecoderModule
from .config import LlamaConfig

ATTENTION_DROPOUT = "attention_probs_dropout_prob"
HIDDEN_DROPOUT = "hidden_dropout_prob"
EXTRA_KWARG_KEYS = [ATTENTION_DROPOUT, HIDDEN_DROPOUT]

# Order-dependent.
COMMON_HF_PARAM_KEY_TRANSFORMS: List[StringTransform] = [
    # Attention blocks.
    StringTransformations.sub(".self_attn", ".mha"),
    StringTransformations.sub(".q_proj", ".query"),
    StringTransformations.sub(".k_proj", ".key"),
    StringTransformations.sub(".v_proj", ".value"),
    StringTransformations.sub(".o_proj", ".output"),
    # Pointwise feedforward
    StringTransformations.sub(".mlp", ".ffn"),
    StringTransformations.sub(".up_proj", ".intermediate"),
    StringTransformations.sub("ffn.down_proj", "ffn.output"),
    StringTransformations.sub(".gate_proj", ".gate"),
    # RMS norms
    StringTransformations.sub(".input_layernorm", ".attn_input_layer_norm"),
    StringTransformations.sub(".post_attention_layernorm", ".ffn_input_layer_norm"),
    StringTransformations.regex_sub(
        (r"^(decoder\.)?norm\.", "\\1output_layer_norm."),
        (r"^(decoder\.)?output_layer_norm\.", "\\1norm."),
    ),
    # Embeddings
    StringTransformations.sub("embed_tokens.", "embeddings.piece_embeddings."),
    StringTransformations.sub("lm_head.", "output_embeddings."),
]

DECODER_HF_PARAM_KEY_TRANSFORMS = [
    StringTransformations.remove_prefix("model.", reversible=False)
] + COMMON_HF_PARAM_KEY_TRANSFORMS
CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS = [
    StringTransformations.sub("model.", "decoder.")
] + COMMON_HF_PARAM_KEY_TRANSFORMS


class HFConfigKeys:
    @staticmethod
    def conv_n_attention_query_heads(config: LlamaConfig) -> float:
        return config.layer.attention.n_query_heads

    @staticmethod
    def conv_n_attention_keyvalue_heads(config: LlamaConfig) -> float:
        return config.layer.attention.n_key_value_heads

    NUM_ATTENTION_HEADS = HFConfigKey(
        "num_attention_heads",
        "n_query_heads",
        lambda c: HFConfigKeys.conv_n_attention_query_heads(c),
    )
    RMS_NORM_EPS = HFConfigKey(
        "rms_norm_eps",
        "rms_norm_eps",
        lambda c: CommonCuratedToHFConverters.layer_norm_eps(c),
    )
    NUM_KEY_VALUE_HEADS = HFConfigKey(
        "num_key_value_heads",
        "n_key_value_heads",
        lambda c: HFConfigKeys.conv_n_attention_keyvalue_heads(c),
    )


HF_CONFIG_KEYS: List[Tuple[HFConfigKey, Optional[HFConfigKeyDefault]]] = [
    (CommonHFKeys.DTYPE, HFConfigKeyDefault("float16")),
    (CommonHFKeys.HIDDEN_ACT, None),
    (CommonHFKeys.HIDDEN_SIZE, None),
    (CommonHFKeys.INTERMEDIATE_SIZE, None),
    (CommonHFKeys.VOCAB_SIZE, None),
    (CommonHFKeys.NUM_HIDDEN_LAYERS, None),
    (HFConfigKeys.NUM_ATTENTION_HEADS, None),
    (HFConfigKeys.RMS_NORM_EPS, None),
    (HFConfigKeys.NUM_KEY_VALUE_HEADS, HFConfigKeyDefault(-1)),
    (CommonHFKeys.ATTENTION_PROBS_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
    (CommonHFKeys.HIDDEN_DROPOUT_PROB, HFConfigKeyDefault(0.0)),
]

HF_SPECIFIC_CONFIG_DECODER = HFSpecificConfig(
    architectures=["LlamaModel"], model_type="llama"
)
HF_SPECIFIC_CONFIG_CAUSAL_LM = HFSpecificConfig(
    architectures=["LlamaForCausalLM"], model_type="llama"
)


def _config_from_hf(hf_config: Mapping[str, Any]) -> LlamaConfig:
    kwargs = config_from_hf("Llama", hf_config, HF_CONFIG_KEYS)

    n_kv_heads = HFConfigKeys.NUM_KEY_VALUE_HEADS.get_kwarg(kwargs)
    if n_kv_heads == -1:
        HFConfigKeys.NUM_KEY_VALUE_HEADS.set_kwarg(
            HFConfigKeys.NUM_ATTENTION_HEADS, kwargs
        )

    return LlamaConfig(
        rotary_embedding_base=10000, rotary_embedding_fraction=1.0, **kwargs
    )


def _config_to_hf(cls, curated_config: LlamaConfig) -> Dict[str, Any]:
    out = config_to_hf(curated_config, [k for k, _ in HF_CONFIG_KEYS])
    if issubclass(cls, DecoderModule):
        return HF_SPECIFIC_CONFIG_DECODER.merge(out)
    else:
        return HF_SPECIFIC_CONFIG_CAUSAL_LM.merge(out)



from typing import Generic, List, Optional

import torch
from torch import Tensor
from torch.nn import Module, ModuleList

from ..layers.attention import AttentionMask
from ..layers.cache import KeyValueCache
from .module import CausalLMModule, ConfigT, DecoderModule, EncoderModule
from .output import CausalLMOutputWithCache, ModelOutput, ModelOutputWithCache


class TransformerDecoder(Generic[ConfigT], DecoderModule[ConfigT, KeyValueCache]):
    """
    Transformer decoder (`Vaswani et al., 2017`_) base class.

    This class provides an implementation of the ``forward`` method.
    Subclasses must set the given member attributes.

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    embeddings: Module
    layers: ModuleList
    output_layer_norm: Module

    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        *,
        cache: Optional[List[KeyValueCache]] = None,
        positions: Optional[Tensor] = None,
        store_cache: bool = False,
    ) -> ModelOutputWithCache[KeyValueCache]:
        embeddings = self.embeddings(piece_ids)
        layer_output = embeddings

        layer_outputs = []
        new_cache = []
        layer_cache = None
        for layer in self.layers:
            if cache is not None:
                # The key-value cache is stored per layer, so peel off one
                # layer at a time.
                layer_cache = cache[0]
                cache = cache[1:]
            layer_output, new_layer_cache = layer(
                layer_output,
                attention_mask,
                cache=layer_cache,
                store_cache=store_cache,
                positions=positions,
            )
            layer_outputs.append(layer_output)
            if store_cache:
                new_cache.append(new_layer_cache)

        layer_outputs[-1] = self.output_layer_norm(layer_outputs[-1])

        return ModelOutputWithCache(
            all_outputs=[embeddings, *layer_outputs],
            cache=new_cache if store_cache else None,
        )


class TransformerCausalLM(Generic[ConfigT], CausalLMModule[ConfigT, KeyValueCache]):
    """
    Transformer causal LM (`Vaswani et al., 2017`_) base class.

    This class provides an implementation of the ``forward`` method.
    Subclasses must set the given member attributes..

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    decoder: TransformerDecoder
    output_embeddings: Module

    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        cache: Optional[List[KeyValueCache]] = None,
        positions: Optional[Tensor] = None,
        store_cache: bool = False,
    ) -> CausalLMOutputWithCache[KeyValueCache]:
        decoder_output = self.decoder(
            piece_ids,
            attention_mask,
            cache=cache,
            store_cache=store_cache,
            positions=positions,
        )
        logits = self.output_embeddings(decoder_output.last_hidden_layer_state)
        return CausalLMOutputWithCache(
            all_outputs=decoder_output.all_outputs,
            cache=decoder_output.cache,
            logits=logits,
        )


class TransformerEncoder(Generic[ConfigT], EncoderModule[ConfigT]):
    """
    Transformer encoder (`Vaswani et al., 2017`_) base class.

    This class provides an implementation of the ``forward`` method.
    Subclasses must set the given member attributes.

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    """

    embeddings: Module
    layers: ModuleList

    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        *,
        positions: Optional[Tensor] = None,
        type_ids: Optional[Tensor] = None,
    ) -> ModelOutput:
        embeddings = self.embeddings(piece_ids, positions=positions, type_ids=type_ids)
        layer_output = embeddings

        layer_outputs = []
        for layer in self.layers:
            layer_output, _ = layer(layer_output, attention_mask)
            layer_outputs.append(layer_output)

        return ModelOutput(all_outputs=[embeddings, *layer_outputs])



import unicodedata
from abc import ABC, abstractmethod
from enum import Enum
from typing import Iterable, List, Optional, Union

from ..chunks import InputChunks, MergedInputChunks, SpecialPieceChunk, TextChunk
from ..tokenizer import PiecesWithIds, TokenizerBase


class PreDecoder(ABC):
    """
    Callable applied before decoding.
    """

    @abstractmethod
    def __call__(self, input: Iterable[Iterable[int]]) -> List[List[int]]:
        """
        Apply the pre-decoder on the input.

        :param input:
            Piece identifiers of each input sequence.
        :returns:
            Modified piece identifiers.
        """
        ...


class PostDecoder(ABC):
    """
    Callable applied after decoding.
    """

    @abstractmethod
    def __call__(self, output: Iterable[str]) -> List[str]:
        """
        Apply the post-decoder on the output.

        :param output:
            Decoded strings from the tokenizer.
        :returns:
            Modified decoded strings.
        """
        ...


class PreEncoder(ABC):
    """
    Callable applied before encoding.
    """

    @abstractmethod
    def __call__(self, chunks: Iterable[InputChunks]) -> List[InputChunks]:
        """
        Apply the pre-encoder on the chunks.

        :param chunks:
            Input chunks of each input sequence.
        :returns:
            Modified input chunks.
        """
        ...


class PostEncoder(ABC):
    """
    Callable applied after encoding.
    """

    @abstractmethod
    def __call__(self, pieces: PiecesWithIds) -> PiecesWithIds:
        """
        Apply the post-encoder on the pieces.

        :param pieces:
            Encoded output of the tokenzier.
        :returns:
            Modified encoded output.
        """
        ...


class Normalizer(ABC):
    """
    Callable applied before encoding.
    """

    @abstractmethod
    def __call__(self, chunks: Iterable[InputChunks]) -> List[InputChunks]:
        """
        Apply the normalizer on the chunks.

        :param chunks:
            Input chunks of each input sequence.
        :returns:
            Modified input chunks.
        """
        ...


class LegacyTokenizer(TokenizerBase):
    """
    Base class for legacy tokenizers.
    """

    normalizer: Optional[Normalizer] = None
    pre_decoder: Optional[PreDecoder] = None
    post_decoder: Optional[PostDecoder] = None
    pre_encoder: Optional[PreEncoder] = None
    post_encoder: Optional[PostEncoder] = None

    def __call__(
        self, input: Union[Iterable[InputChunks], Iterable[str]]
    ) -> PiecesWithIds:
        return self.encode(input)

    def decode(
        self, input: Iterable[Iterable[int]], skip_special_pieces: bool = True
    ) -> List[str]:
        if self.pre_decoder is not None:
            input = self.pre_decoder(input)

        strings = self._decode(input, skip_special_pieces)

        if self.post_decoder is not None:
            strings = self.post_decoder(strings)

        return strings

    def encode(
        self, input: Union[Iterable[InputChunks], Iterable[str]]
    ) -> PiecesWithIds:
        chunks = self._convert_strings(input)

        if self.normalizer is not None:
            chunks = self.normalizer(chunks)
        if self.pre_encoder is not None:
            chunks = self.pre_encoder(chunks)

        merged_chunks = [seq_chunks.merge_text_chunks() for seq_chunks in chunks]
        pieces = self._encode(merged_chunks)

        if self.post_encoder is not None:
            pieces = self.post_encoder(pieces)

        return pieces

    def _convert_strings(
        self, input: Union[Iterable[InputChunks], Iterable[str]]
    ) -> List[InputChunks]:
        return [
            InputChunks([TextChunk(seq)]) if isinstance(seq, str) else seq
            for seq in input
        ]

    @abstractmethod
    def _decode(
        self, input: Iterable[Iterable[int]], skip_special_pieces: bool
    ) -> List[str]: ...

    @abstractmethod
    def _encode(self, input: Iterable[MergedInputChunks]) -> PiecesWithIds: ...


class AddBosEosPreEncoder(PreEncoder):
    """
    Adds beginning/end of sequence markers before the encoding process.
    """

    def __init__(
        self,
        *,
        bos_piece: Optional[str],
        eos_piece: Optional[str],
    ):
        """
        Construct a pre-encoder that adds beginning/end of sequence markers.

        :param bos_piece:
            The piece used to mark the beginning of a sequence. The piece
            is not added when its value is ``None``.
        :param eos_piece:
            The piece used to mark the end of a sequence. The piece is not
            added when its value is ``None``.
        """
        self.bos_piece = bos_piece
        self.eos_piece = eos_piece

    def __call__(self, chunks: Iterable[InputChunks]) -> List[InputChunks]:
        bos_eos_chunks = []
        for seq_chunks in chunks:
            bos_chunks = (
                [SpecialPieceChunk(self.bos_piece)]
                if self.bos_piece is not None
                else []
            )
            eos_chunks = (
                [SpecialPieceChunk(self.eos_piece)]
                if self.eos_piece is not None
                else []
            )
            bos_eos_chunks.append(InputChunks([*bos_chunks, *seq_chunks, *eos_chunks]))
        return bos_eos_chunks


class UnicodeNormalization(str, Enum):
    """
    Unicode normalization schemes.
    """

    NFC = "NFC"
    NFKC = "NFKC"
    NFD = "NFD"
    NFKD = "NFKD"


class DefaultNormalizer(Normalizer):
    """
    Performs normalization operations on input chunks before encoding.
    """

    def __init__(
        self,
        *,
        utf_normalization: Optional[UnicodeNormalization] = None,
        lowercase: bool = False,
        strip_accents: bool = False,
    ) -> None:
        """
        Construct a default normalizer.

        :param utf_normalization:
            Unicode normalization scheme to use.
        :param lowercase:
            Lowercase text.
        :param strip_accents:
            Remove accents from text.
        """
        super().__init__()
        self.utf_normalization = utf_normalization
        self.lowercase = lowercase
        self.strip_accents = strip_accents

    def __call__(self, chunks: Iterable[InputChunks]) -> List[InputChunks]:
        chunks = list(chunks)
        for chunk in chunks:
            for piece_or_text in chunk:
                if not isinstance(piece_or_text, TextChunk):
                    continue
                text = piece_or_text.text
                if self.lowercase:
                    text = text.lower()
                if self.strip_accents:
                    # Normalize with NFD to decompose accents.
                    text = unicodedata.normalize(UnicodeNormalization.NFD, text)
                    text = "".join(
                        [char for char in text if unicodedata.category(char) != "Mn"]
                    )
                if self.utf_normalization is not None:
                    # This has to be done at the end to ensure that previously
                    # applied normalization is correctly overridden.
                    text = unicodedata.normalize(self.utf_normalization, text)
                piece_or_text.text = text
        return chunks



from dataclasses import dataclass
from typing import Optional, Protocol, Tuple, Type, TypeVar, Union

import torch
from torch import Tensor

CacheProtocolSelf = TypeVar("CacheProtocolSelf", bound="CacheProtocol")


class CacheProtocol(Protocol):
    def filter_batch_items(self: CacheProtocolSelf, mask: Tensor) -> CacheProtocolSelf:
        """
        Filter batch sequences from the cache.

        Sequences for which the mask is ``True`` are retained.

        :param mask:
            Mask of batch items to retain.

            *Shape:* ``(batch_size,)``
        :returns:
            Filtered items.
        """
        ...


@dataclass
class KeyValueCache:
    """
    Cache type for layers that cache keys and values.

    :param key:
        Key.
    :param value:
        Value.
    """

    key: Tensor
    value: Tensor

    def filter_batch_items(self, mask: Tensor) -> "KeyValueCache":
        if mask.ndim != 1:
            raise ValueError(
                f"Cache mask must be a 1D tensor, has {mask.ndim} dimensions."
            )
        if mask.size(0) != self.key.size(0):
            raise ValueError(
                f"Cache mask size ({mask.size(0)}) must match cache batch size ({self.key.size(0)})."
            )
        if mask.dtype != torch.bool:
            raise ValueError(f"Cache mask dtype must be bool, was: {mask.dtype}.")

        return KeyValueCache(key=self.key[mask], value=self.value[mask])



from functools import partial
from typing import Optional

import torch
from torch import Tensor
from torch.nn import LayerNorm, Module, ModuleList

from ...layers.attention import (
    AttentionHeads,
    AttentionMask,
    QkvMode,
    QkvSplitGroupedByKVHeads,
    ScaledDotProductAttention,
    SelfAttention,
)
from ...layers.feedforward import PointwiseFeedForward
from ...layers.transformer import (
    EncoderLayer,
    TransformerDropouts,
    TransformerLayerNorms,
)
from .config import ALBERTLayerConfig


class ALBERTLayerGroup(Module):
    """
    ALBERT (`Lan et al., 2022`_) layer group.

    .. _Lan et al., 2022: https://arxiv.org/abs/1909.11942
    """

    def __init__(
        self, layer_config: ALBERTLayerConfig, *, device: Optional[torch.device] = None
    ) -> None:
        super().__init__()

        layer_norm = partial(
            LayerNorm,
            layer_config.feedforward.hidden_width,
            layer_config.layer_norm_eps,
            device=device,
        )
        attention_config = layer_config.attention
        self.group_layers = ModuleList(
            [
                EncoderLayer(
                    attention_layer=SelfAttention(
                        attention_heads=AttentionHeads.uniform(
                            attention_config.n_query_heads,
                            QkvSplitGroupedByKVHeads(),
                        ),
                        attention_scorer=ScaledDotProductAttention(
                            dropout_prob=attention_config.dropout_prob,
                            linear_biases=None,
                        ),
                        hidden_width=layer_config.feedforward.hidden_width,
                        qkv_mode=QkvMode.SEPARATE,
                        rotary_embeds=None,
                        use_bias=attention_config.use_bias,
                        device=device,
                    ),
                    feed_forward_layer=PointwiseFeedForward(
                        activation=layer_config.feedforward.activation.module(),
                        hidden_width=layer_config.feedforward.hidden_width,
                        intermediate_width=layer_config.feedforward.intermediate_width,
                        use_bias=layer_config.feedforward.use_bias,
                        use_gate=layer_config.feedforward.use_gate,
                        device=device,
                    ),
                    dropouts=TransformerDropouts.layer_output_dropouts(
                        layer_config.dropout_prob
                    ),
                    layer_norms=TransformerLayerNorms(
                        attn_residual_layer_norm=layer_norm(),
                        ffn_residual_layer_norm=layer_norm(),
                    ),
                    use_parallel_attention=attention_config.use_parallel_attention,
                )
                for _ in range(layer_config.n_layers_per_group)
            ]
        )

    def forward(self, input: Tensor, attention_mask: AttentionMask) -> Tensor:
        """
        Apply the ALBERT layer group to the given piece hidden representations.

        :param input:
            Hidden representations to apply the layer group to.

            *Shape:* ``(batch_size, seq_len, width)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            during attention calculation.
        :returns:
            Layer output.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        layer_output = input
        for layer in self.group_layers:
            layer_output = layer(layer_output, attention_mask)
        return layer_output



from pathlib import Path

import pytest
import torch

TORCH_DEVICES = [torch.device("cpu")]
GPU_TESTS_ENABLED = False


def pytest_addoption(parser):
    try:
        parser.addoption("--slow", action="store_true", help="include slow tests")
        parser.addoption(
            "--upload-tests",
            action="store_true",
            help="include tests that upload test artifacts to remote repos",
        )
    # Options are already added, e.g. if conftest is copied in a build pipeline
    # and runs twice
    except ValueError:
        pass


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: include slow tests")
    config.addinivalue_line(
        "markers", "upload-tests: tests that upload test artifacts to remote repos"
    )


def pytest_runtest_setup(item):
    def getopt(opt):
        # When using 'pytest --pyargs spacy' to test an installed copy of
        # spacy, pytest skips running our pytest_addoption() hook. Later, when
        # we call getoption(), pytest raises an error, because it doesn't
        # recognize the option we're asking about. To avoid this, we need to
        # pass a default value. We default to False, i.e., we act like all the
        # options weren't given.
        return item.config.getoption(f"--{opt}", False)

    # Integration of boolean flags
    for opt in ["upload_tests", "slow"]:
        if opt in item.keywords and not getopt(opt.replace("_", "-")):
            pytest.skip(f"need --{opt.replace('_', '-')} option to run")


@pytest.fixture
def test_dir(request):
    return Path(request.fspath).parent


@pytest.fixture
def french_sample_texts():
    return [
        "J'ai vu une fille avec un télescope.",
        "Aujourd'hui, nous allons manger un poké bowl.",
    ]


@pytest.fixture
def short_sample_texts():
    return [
        "I saw a girl with a telescope.",
        "Today we will eat poké bowl, lots of it!",
        "Tokens which are unknown inペ mostで latinが alphabet際 vocabularies.",
    ]


@pytest.fixture
def sample_texts():
    # Two short Wikipedia fragments from:
    # https://en.wikipedia.org/wiki/Kinesis_(keyboard)#Contoured_/_Advantage
    # https://en.wikipedia.org/wiki/Doom_(1993_video_game)#Engine
    return [
        "The original Model 100, released in 1992, featured a single-piece "
        "contoured design similar to the Maltron keyboard, with the keys laid "
        "out in a traditional QWERTY arrangement, separated into two clusters "
        "for the left and right hands.[2] A 1993 article in PC Magazine "
        "described the US$690 (equivalent to $1,300 in 2021) keyboard's "
        'arrangement as having "the alphabet keys in precisely vertical '
        "(not diagonal) columns in two concave depressions. The Kinesis "
        "Keyboard also puts the Backspace, Delete, Enter, Space, Ctrl, Alt, "
        "Home, End, Page Up, and Page Down keys under your thumbs in the "
        'middle".[23]',
        "Doom was programmed largely in the ANSI C programming language, with "
        "a few elements in assembly language. Development was done on NeXT "
        "computers running the NeXTSTEP operating system.[35] The data used by "
        "the game engine, including level designs and graphics files, are "
        'stored in WAD files, short for "Where\'s All the Data?".',
    ]
